# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/arith.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void arith_primorial(fmpz_t res, slong n) noexcept
    void _arith_harmonic_number(fmpz_t num, fmpz_t den, slong n) noexcept
    void arith_harmonic_number(fmpq_t x, slong n) noexcept
    void arith_stirling_number_1u(fmpz_t s, ulong n, ulong k) noexcept
    void arith_stirling_number_1(fmpz_t s, ulong n, ulong k) noexcept
    void arith_stirling_number_2(fmpz_t s, ulong n, ulong k) noexcept
    void arith_stirling_number_1u_vec(fmpz * row, ulong n, slong klen) noexcept
    void arith_stirling_number_1_vec(fmpz * row, ulong n, slong klen) noexcept
    void arith_stirling_number_2_vec(fmpz * row, ulong n, slong klen) noexcept
    void arith_stirling_number_1u_vec_next(fmpz * row, const fmpz * prev, slong n, slong klen) noexcept
    void arith_stirling_number_1_vec_next(fmpz * row, const fmpz * prev, slong n, slong klen) noexcept
    void arith_stirling_number_2_vec_next(fmpz * row, const fmpz * prev, slong n, slong klen) noexcept
    void arith_stirling_matrix_1u(fmpz_mat_t mat) noexcept
    void arith_stirling_matrix_1(fmpz_mat_t mat) noexcept
    void arith_stirling_matrix_2(fmpz_mat_t mat) noexcept
    void arith_bell_number(fmpz_t b, ulong n) noexcept
    void arith_bell_number_dobinski(fmpz_t res, ulong n) noexcept
    void arith_bell_number_multi_mod(fmpz_t res, ulong n) noexcept
    void arith_bell_number_vec(fmpz * b, slong n) noexcept
    void arith_bell_number_vec_recursive(fmpz * b, slong n) noexcept
    void arith_bell_number_vec_multi_mod(fmpz * b, slong n) noexcept
    mp_limb_t arith_bell_number_nmod(ulong n, nmod_t mod) noexcept
    void arith_bell_number_nmod_vec(mp_ptr b, slong n, nmod_t mod) noexcept
    void arith_bell_number_nmod_vec_recursive(mp_ptr b, slong n, nmod_t mod) noexcept
    void arith_bell_number_nmod_vec_ogf(mp_ptr b, slong n, nmod_t mod) noexcept
    int arith_bell_number_nmod_vec_series(mp_ptr b, slong n, nmod_t mod) noexcept
    double arith_bell_number_size(ulong n) noexcept
    void _arith_bernoulli_number(fmpz_t num, fmpz_t den, ulong n) noexcept
    void arith_bernoulli_number(fmpq_t x, ulong n) noexcept
    void _arith_bernoulli_number_vec(fmpz * num, fmpz * den, slong n) noexcept
    void arith_bernoulli_number_vec(fmpq * x, slong n) noexcept
    void arith_bernoulli_number_denom(fmpz_t den, ulong n) noexcept
    double arith_bernoulli_number_size(ulong n) noexcept
    void arith_bernoulli_polynomial(fmpq_poly_t poly, ulong n) noexcept
    void _arith_bernoulli_number_vec_recursive(fmpz * num, fmpz * den, slong n) noexcept
    void _arith_bernoulli_number_vec_multi_mod(fmpz * num, fmpz * den, slong n) noexcept
    void arith_euler_number(fmpz_t res, ulong n) noexcept
    void arith_euler_number_vec(fmpz * res, slong n) noexcept
    double arith_euler_number_size(ulong n) noexcept
    void arith_euler_polynomial(fmpq_poly_t poly, ulong n) noexcept
    void arith_euler_phi(fmpz_t res, const fmpz_t n) noexcept
    int arith_moebius_mu(const fmpz_t n) noexcept
    void arith_divisor_sigma(fmpz_t res, ulong k, const fmpz_t n) noexcept
    void arith_divisors(fmpz_poly_t res, const fmpz_t n) noexcept
    void arith_ramanujan_tau(fmpz_t res, const fmpz_t n) noexcept
    void arith_ramanujan_tau_series(fmpz_poly_t res, slong n) noexcept
    void arith_landau_function_vec(fmpz * res, slong len) noexcept
    void arith_dedekind_sum_naive(fmpq_t s, const fmpz_t h, const fmpz_t k) noexcept
    double arith_dedekind_sum_coprime_d(double h, double k) noexcept
    void arith_dedekind_sum_coprime_large(fmpq_t s, const fmpz_t h, const fmpz_t k) noexcept
    void arith_dedekind_sum_coprime(fmpq_t s, const fmpz_t h, const fmpz_t k) noexcept
    void arith_dedekind_sum(fmpq_t s, const fmpz_t h, const fmpz_t k) noexcept
    void arith_number_of_partitions_vec(fmpz * res, slong len) noexcept
    void arith_number_of_partitions_nmod_vec(mp_ptr res, slong len, nmod_t mod) noexcept
    void arith_hrr_expsum_factored(trig_prod_t prod, mp_limb_t k, mp_limb_t n) noexcept
    void arith_number_of_partitions_mpfr(mpfr_t x, ulong n) noexcept
    void arith_number_of_partitions(fmpz_t x, ulong n) noexcept
    void arith_sum_of_squares(fmpz_t r, ulong k, const fmpz_t n) noexcept
    void arith_sum_of_squares_vec(fmpz * r, ulong k, slong n) noexcept



# sage_setup: distribution = sagemath-schemes
"""
Plane curves
"""

# *****************************************************************************
#
#   Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.schemes.curves.constructor import Curve
from sage.schemes.curves.projective_curve import Hasse_bounds

from sage.misc.lazy_import import lazy_import

lazy_import('sage.schemes.curves.plane_curve_arrangement', 'PlaneCurveArrangements')

lazy_import('sage.schemes.curves.plane_curve_arrangement', 'AffinePlaneCurveArrangements')

lazy_import('sage.schemes.curves.plane_curve_arrangement', 'ProjectivePlaneCurveArrangements')



# sage_setup: distribution = sagemath-ntl
# distutils: libraries = gmp mpfr mpfi

from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.mpfi.types cimport *

cdef extern from "mpfi.h":
    # Rounding
    int mpfi_round_prec(mpfi_ptr, mpfr_prec_t prec)

    # Initialization, destruction, and assignment
    # Initializations
    void mpfi_init(mpfi_ptr)
    void mpfi_init2(mpfi_ptr, mpfr_prec_t)
    void mpfi_clear(mpfi_ptr)

    # mpfi bounds have the same precision
    mpfr_prec_t mpfi_get_prec(mpfi_srcptr)
    void mpfi_set_prec(mpfi_ptr, mpfr_prec_t)

    # assignment functions
    int mpfi_set(mpfi_ptr, mpfi_srcptr)
    int mpfi_set_si(mpfi_ptr, long)
    int mpfi_set_ui(mpfi_ptr, unsigned long)
    int mpfi_set_d(mpfi_ptr, double)
    int mpfi_set_z(mpfi_ptr, mpz_t)
    int mpfi_set_q(mpfi_ptr, mpq_t)
    int mpfi_set_fr(mpfi_ptr, mpfr_srcptr)
    int mpfi_set_str(mpfi_ptr, const char *, int)

    # combined initialization and assignment functions
    int mpfi_init_set(mpfi_ptr, mpfi_srcptr)
    int mpfi_init_set_si(mpfi_ptr, long)
    int mpfi_init_set_ui(mpfi_ptr, unsigned long)
    int mpfi_init_set_d(mpfi_ptr, double)
    int mpfi_init_set_z(mpfi_ptr, mpz_srcptr)
    int mpfi_init_set_q(mpfi_ptr, mpq_srcptr)
    int mpfi_init_set_fr(mpfi_ptr, mpfr_srcptr)
    int mpfi_init_set_str(mpfi_ptr, const char *, int)

    # swapping two intervals
    void mpfi_swap(mpfi_ptr, mpfi_ptr)

    # Various useful interval functions
    # with scalar or interval results

    # absolute diameter
    int mpfi_diam_abs(mpfr_ptr, mpfi_srcptr)
    # relative diameter
    int mpfi_diam_rel(mpfr_ptr, mpfi_srcptr)
    # diameter: relative if the interval does not contain 0
    # absolute otherwise
    int mpfi_diam(mpfr_ptr, mpfi_srcptr)
    # magnitude: the largest absolute value of any element
    int mpfi_mag(mpfr_ptr, mpfi_srcptr)
    # magnitude: the smallest absolute value of any element
    int mpfi_mig(mpfr_ptr, mpfi_srcptr)
    # middle of y
    int mpfi_mid(mpfr_ptr, mpfi_srcptr)
    # picks randomly a point m in y
    void mpfi_alea(mpfr_ptr, mpfi_srcptr)

    # Conversions
    double mpfi_get_d(mpfi_srcptr)
    void mpfi_get_fr(mpfr_ptr, mpfi_srcptr)

    # Basic arithmetic operations

    # arithmetic operations between two interval operands
    int mpfi_add(mpfi_ptr, mpfi_srcptr, mpfi_srcptr)
    int mpfi_sub(mpfi_ptr, mpfi_srcptr, mpfi_srcptr)
    int mpfi_mul(mpfi_ptr, mpfi_srcptr, mpfi_srcptr)
    int mpfi_div(mpfi_ptr, mpfi_srcptr, mpfi_srcptr)

    # arithmetic operations between an interval operand and a double prec. floating-point
    int mpfi_add_d(mpfi_ptr, mpfi_srcptr, double)
    int mpfi_sub_d(mpfi_ptr, mpfi_srcptr, double)
    int mpfi_d_sub(mpfi_ptr, double, mpfi_srcptr)
    int mpfi_mul_d(mpfi_ptr, mpfi_srcptr, double)
    int mpfi_div_d(mpfi_ptr, mpfi_srcptr, double)
    int mpfi_d_div(mpfi_ptr, double, mpfi_srcptr)

    # arithmetic operations between an interval operand and an unsigned long integer
    int mpfi_add_ui(mpfi_ptr, mpfi_srcptr, unsigned long)
    int mpfi_sub_ui(mpfi_ptr, mpfi_srcptr, unsigned long)
    int mpfi_ui_sub(mpfi_ptr, unsigned long, mpfi_srcptr)
    int mpfi_mul_ui(mpfi_ptr, mpfi_srcptr, unsigned long)
    int mpfi_div_ui(mpfi_ptr, mpfi_srcptr, unsigned long)
    int mpfi_ui_div(mpfi_ptr, unsigned long, mpfi_srcptr)

    # arithmetic operations between an interval operand and a long integer
    int mpfi_add_si(mpfi_ptr, mpfi_srcptr, long)
    int mpfi_sub_si(mpfi_ptr, mpfi_srcptr, long)
    int mpfi_si_sub(mpfi_ptr, long, mpfi_srcptr)
    int mpfi_mul_si(mpfi_ptr, mpfi_srcptr, long)
    int mpfi_div_si(mpfi_ptr, mpfi_srcptr, long)
    int mpfi_si_div(mpfi_ptr, long, mpfi_srcptr)

    # arithmetic operations between an interval operand and a multiple prec. integer
    int mpfi_add_z(mpfi_ptr, mpfi_srcptr, mpz_srcptr)
    int mpfi_sub_z(mpfi_ptr, mpfi_srcptr, mpz_srcptr)
    int mpfi_z_sub(mpfi_ptr, mpz_srcptr, mpfi_srcptr)
    int mpfi_mul_z(mpfi_ptr, mpfi_srcptr, mpz_srcptr)
    int mpfi_div_z(mpfi_ptr, mpfi_srcptr, mpz_srcptr)
    int mpfi_z_div(mpfi_ptr, mpz_srcptr, mpfi_srcptr)

    # arithmetic operations between an interval operand and a multiple prec. rational
    int mpfi_add_q(mpfi_ptr, mpfi_srcptr, mpq_srcptr)
    int mpfi_sub_q(mpfi_ptr, mpfi_srcptr, mpq_srcptr)
    int mpfi_q_sub(mpfi_ptr, mpq_srcptr, mpfi_srcptr)
    int mpfi_mul_q(mpfi_ptr, mpfi_srcptr, mpq_srcptr)
    int mpfi_div_q(mpfi_ptr, mpfi_srcptr, mpq_srcptr)
    int mpfi_q_div(mpfi_ptr, mpq_srcptr, mpfi_srcptr)

    # arithmetic operations between an interval operand and a mult. prec. floating-pt nb
    int mpfi_add_fr(mpfi_ptr, mpfi_srcptr, mpfr_srcptr)
    int mpfi_sub_fr(mpfi_ptr, mpfi_srcptr, mpfr_srcptr)
    int mpfi_fr_sub(mpfi_ptr, mpfr_srcptr, mpfi_srcptr)
    int mpfi_mul_fr(mpfi_ptr, mpfi_srcptr, mpfr_srcptr)
    int mpfi_div_fr(mpfi_ptr, mpfi_srcptr, mpfr_srcptr)
    int mpfi_fr_div(mpfi_ptr, mpfr_srcptr, mpfi_srcptr)

    # arithmetic operations taking a single interval operand
    int mpfi_neg(mpfi_ptr, mpfi_srcptr)
    int mpfi_sqr(mpfi_ptr, mpfi_srcptr)
    # the inv function generates the whole real interval
    # if 0 is in the interval defining the divisor
    int mpfi_inv(mpfi_ptr, mpfi_srcptr)
    # the sqrt of a (partially) negative interval is a NaN
    int mpfi_sqrt(mpfi_ptr, mpfi_srcptr)
    # the first interval contains the absolute values of
    # every element of the second interval
    int mpfi_abs(mpfi_ptr, mpfi_srcptr)

    # various operations
    int mpfi_mul_2exp(mpfi_ptr, mpfi_srcptr, unsigned long)
    int mpfi_mul_2ui(mpfi_ptr, mpfi_srcptr, unsigned long)
    int mpfi_mul_2si(mpfi_ptr, mpfi_srcptr, long)
    int mpfi_div_2exp(mpfi_ptr, mpfi_srcptr, unsigned long)
    int mpfi_div_2ui(mpfi_ptr, mpfi_srcptr, unsigned long)
    int mpfi_div_2si(mpfi_ptr, mpfi_srcptr, long)

    # Special functions
    int mpfi_log(mpfi_ptr, mpfi_srcptr)
    int mpfi_exp(mpfi_ptr, mpfi_srcptr)
    int mpfi_exp2(mpfi_ptr, mpfi_srcptr)

    int mpfi_cos(mpfi_ptr, mpfi_srcptr)
    int mpfi_sin(mpfi_ptr, mpfi_srcptr)
    int mpfi_tan(mpfi_ptr, mpfi_srcptr)
    int mpfi_acos(mpfi_ptr, mpfi_srcptr)
    int mpfi_asin(mpfi_ptr, mpfi_srcptr)
    int mpfi_atan(mpfi_ptr, mpfi_srcptr)

    int mpfi_cosh(mpfi_ptr, mpfi_srcptr)
    int mpfi_sinh(mpfi_ptr, mpfi_srcptr)
    int mpfi_tanh(mpfi_ptr, mpfi_srcptr)
    int mpfi_acosh(mpfi_ptr, mpfi_srcptr)
    int mpfi_asinh(mpfi_ptr, mpfi_srcptr)
    int mpfi_atanh(mpfi_ptr, mpfi_srcptr)

    int mpfi_log1p(mpfi_ptr, mpfi_srcptr)
    int mpfi_expm1(mpfi_ptr, mpfi_srcptr)

    int mpfi_log2(mpfi_ptr, mpfi_srcptr)
    int mpfi_log10(mpfi_ptr, mpfi_srcptr)

    int mpfi_const_log2(mpfi_ptr)
    int mpfi_const_pi(mpfi_ptr)
    int mpfi_const_euler(mpfi_ptr)

    # Comparison functions
    # Warning: the meaning of interval comparison is not clearly defined
    # customizable comparison functions

    int (*mpfi_cmp) (mpfi_srcptr, mpfi_srcptr)
    int (*mpfi_cmp_d) (mpfi_srcptr, double)
    int (*mpfi_cmp_ui) (mpfi_srcptr, unsigned long)
    int (*mpfi_cmp_si) (mpfi_srcptr, long)
    int (*mpfi_cmp_z) (mpfi_srcptr, mpz_srcptr)
    int (*mpfi_cmp_q) (mpfi_srcptr, mpq_srcptr)
    int (*mpfi_cmp_fr) (mpfi_srcptr, mpfr_srcptr)

    bint (*mpfi_is_pos) (mpfi_srcptr)
    bint (*mpfi_is_nonneg) (mpfi_srcptr)
    bint (*mpfi_is_neg) (mpfi_srcptr)
    bint (*mpfi_is_nonpos) (mpfi_srcptr)
    bint (*mpfi_is_zero) (mpfi_srcptr)
    bint (*mpfi_is_strictly_pos) (mpfi_srcptr)
    bint (*mpfi_is_strictly_neg) (mpfi_srcptr)

    bint mpfi_has_zero(mpfi_srcptr)

    bint mpfi_nan_p(mpfi_srcptr)
    bint mpfi_inf_p(mpfi_srcptr)
    bint mpfi_bounded_p(mpfi_srcptr)

    # Interval manipulation

    # operations related to the internal representation by endpoints

    # get left or right bound of the interval defined by the
    # second argument and put the result in the first one
    int mpfi_get_left(mpfr_ptr, mpfi_srcptr)
    int mpfi_get_right(mpfr_ptr, mpfi_srcptr)

    int mpfi_revert_if_needed(mpfi_ptr)

    # Set operations on intervals
    # "Convex hulls"
    # extends the interval defined by the first argument
    # so that it contains the second one

    int mpfi_put(mpfi_ptr, mpfi_srcptr)
    int mpfi_put_d(mpfi_ptr, double)
    int mpfi_put_si(mpfi_ptr, long)
    int mpfi_put_ui(mpfi_ptr, unsigned long)
    int mpfi_put_z(mpfi_ptr, mpz_srcptr)
    int mpfi_put_q(mpfi_ptr, mpq_srcptr)
    int mpfi_put_fr(mpfi_ptr, mpfr_srcptr)

    # builds an interval whose left bound is the lower (round -infty)
    # than the second argument and the right bound is greater
    # (round +infty) than the third one


    int mpfi_interv_d(mpfi_ptr, double,double)
    int mpfi_interv_si(mpfi_ptr, long,long)
    int mpfi_interv_ui(mpfi_ptr, unsigned long,unsigned long)
    int mpfi_interv_z(mpfi_ptr, mpz_srcptr,mpz_srcptr)
    int mpfi_interv_q(mpfi_ptr, mpq_srcptr,mpq_srcptr)
    int mpfi_interv_fr(mpfi_ptr, mpfr_srcptr,mpfr_srcptr)

    # Inclusion tests
    # tests if the first argument is inside the interval
    # defined by the second one
    bint mpfi_is_strictly_inside(mpfi_srcptr, mpfi_srcptr)
    bint mpfi_is_inside(mpfi_srcptr, mpfi_srcptr)
    bint mpfi_is_inside_d(double, mpfi_srcptr)
    bint mpfi_is_inside_ui(unsigned long, mpfi_srcptr)
    bint mpfi_is_inside_si(long, mpfi_srcptr)
    bint mpfi_is_inside_z(mpz_srcptr, mpfi_srcptr)
    bint mpfi_is_inside_q(mpq_srcptr, mpfi_srcptr)
    bint mpfi_is_inside_fr(mpfr_srcptr, mpfi_srcptr)

    # set operations
    bint mpfi_is_empty(mpfi_srcptr)
    int mpfi_intersect(mpfi_ptr, mpfi_srcptr, mpfi_srcptr)
    int mpfi_union(mpfi_ptr, mpfi_srcptr, mpfi_srcptr)

    # Miscellaneous

    # adds the second argument to the right bound of the first one
    # and subtracts the second argument to the left bound of
    # the first one
    int mpfi_increase(mpfi_ptr, mpfr_srcptr)
    # keeps the same center and multiply the radius by 2*(1+fact)
    int mpfi_blow(mpfi_ptr, mpfi_srcptr, double)
    # splits the interval into 2 halves
    int mpfi_bisect(mpfi_ptr, mpfi_ptr, mpfi_srcptr)

    char * mpfi_get_version()

    # Error handling
    void mpfi_reset_error()
    void mpfi_set_error(int)
    int mpfi_is_error()

    ctypedef enum mpfi_flags_exact:
        MPFI_FLAGS_BOTH_ENDPOINTS_EXACT
        MPFI_FLAGS_LEFT_ENDPOINT_INEXACT
        MPFI_FLAGS_RIGHT_ENDPOINT_INEXACT
        MPFI_FLAGS_BOTH_ENDPOINTS_INEXACT

    bint MPFI_BOTH_ARE_EXACT(mpfi_flags_exact)
    bint MPFI_LEFT_IS_INEXACT(mpfi_flags_exact)
    bint MPFI_RIGHT_IS_INEXACT(mpfi_flags_exact)
    bint MPFI_BOTH_ARE_INEXACT(mpfi_flags_exact)

    mpfi_flags_exact MPFI_REVERT_INEXACT_FLAGS(mpfi_flags_exact)

    bint MPFI_NAN_P(mpfi_srcptr)
    bint MPFI_INF_P(mpfi_srcptr)
    bint MPFI_IS_ZERO(mpfi_srcptr)

    void MPFI_CLEAR(mpfi_ptr)



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/fmpz_mat.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fmpz_mat_init(fmpz_mat_t mat, slong rows, slong cols) noexcept
    void fmpz_mat_clear(fmpz_mat_t mat) noexcept
    void fmpz_mat_set(fmpz_mat_t mat1, const fmpz_mat_t mat2) noexcept
    void fmpz_mat_init_set(fmpz_mat_t mat, const fmpz_mat_t src) noexcept
    slong fmpz_mat_nrows(const fmpz_mat_t mat) noexcept
    slong fmpz_mat_ncols(const fmpz_mat_t mat) noexcept
    void fmpz_mat_swap(fmpz_mat_t mat1, fmpz_mat_t mat2) noexcept
    void fmpz_mat_swap_entrywise(fmpz_mat_t mat1, fmpz_mat_t mat2) noexcept
    fmpz * fmpz_mat_entry(const fmpz_mat_t mat, slong i, slong j) noexcept
    void fmpz_mat_zero(fmpz_mat_t mat) noexcept
    void fmpz_mat_one(fmpz_mat_t mat) noexcept
    void fmpz_mat_swap_rows(fmpz_mat_t mat, slong * perm, slong r, slong s) noexcept
    void fmpz_mat_swap_cols(fmpz_mat_t mat, slong * perm, slong r, slong s) noexcept
    void fmpz_mat_invert_rows(fmpz_mat_t mat, slong * perm) noexcept
    void fmpz_mat_invert_cols(fmpz_mat_t mat, slong * perm) noexcept
    void fmpz_mat_window_init(fmpz_mat_t window, const fmpz_mat_t mat, slong r1, slong c1, slong r2, slong c2) noexcept
    void fmpz_mat_window_clear(fmpz_mat_t window) noexcept
    void fmpz_mat_randbits(fmpz_mat_t mat, flint_rand_t state, flint_bitcnt_t bits) noexcept
    void fmpz_mat_randtest(fmpz_mat_t mat, flint_rand_t state, flint_bitcnt_t bits) noexcept
    void fmpz_mat_randintrel(fmpz_mat_t mat, flint_rand_t state, flint_bitcnt_t bits) noexcept
    void fmpz_mat_randsimdioph(fmpz_mat_t mat, flint_rand_t state, flint_bitcnt_t bits, flint_bitcnt_t bits2) noexcept
    void fmpz_mat_randntrulike(fmpz_mat_t mat, flint_rand_t state, flint_bitcnt_t bits, ulong q) noexcept
    void fmpz_mat_randntrulike2(fmpz_mat_t mat, flint_rand_t state, flint_bitcnt_t bits, ulong q) noexcept
    void fmpz_mat_randajtai(fmpz_mat_t mat, flint_rand_t state, double alpha) noexcept
    int fmpz_mat_randpermdiag(fmpz_mat_t mat, flint_rand_t state, const fmpz * diag, slong n) noexcept
    void fmpz_mat_randrank(fmpz_mat_t mat, flint_rand_t state, slong rank, flint_bitcnt_t bits) noexcept
    void fmpz_mat_randdet(fmpz_mat_t mat, flint_rand_t state, const fmpz_t det) noexcept
    void fmpz_mat_randops(fmpz_mat_t mat, flint_rand_t state, slong count) noexcept
    int fmpz_mat_fprint(FILE * file, const fmpz_mat_t mat) noexcept
    int fmpz_mat_fprint_pretty(FILE * file, const fmpz_mat_t mat) noexcept
    int fmpz_mat_print(const fmpz_mat_t mat) noexcept
    int fmpz_mat_print_pretty(const fmpz_mat_t mat) noexcept
    int fmpz_mat_fread(FILE * file, fmpz_mat_t mat) noexcept
    int fmpz_mat_read(fmpz_mat_t mat) noexcept
    bint fmpz_mat_equal(const fmpz_mat_t mat1, const fmpz_mat_t mat2) noexcept
    bint fmpz_mat_is_zero(const fmpz_mat_t mat) noexcept
    bint fmpz_mat_is_one(const fmpz_mat_t mat) noexcept
    bint fmpz_mat_is_empty(const fmpz_mat_t mat) noexcept
    bint fmpz_mat_is_square(const fmpz_mat_t mat) noexcept
    bint fmpz_mat_is_zero_row(const fmpz_mat_t mat, slong i) noexcept
    bint fmpz_mat_equal_col(fmpz_mat_t M, slong m, slong n) noexcept
    bint fmpz_mat_equal_row(fmpz_mat_t M, slong m, slong n) noexcept
    void fmpz_mat_transpose(fmpz_mat_t B, const fmpz_mat_t A) noexcept
    void fmpz_mat_concat_vertical(fmpz_mat_t res, const fmpz_mat_t mat1, const fmpz_mat_t mat2) noexcept
    void fmpz_mat_concat_horizontal(fmpz_mat_t res, const fmpz_mat_t mat1, const fmpz_mat_t mat2) noexcept
    void fmpz_mat_get_nmod_mat(nmod_mat_t Amod, const fmpz_mat_t A) noexcept
    void fmpz_mat_set_nmod_mat(fmpz_mat_t A, const nmod_mat_t Amod) noexcept
    void fmpz_mat_set_nmod_mat_unsigned(fmpz_mat_t A, const nmod_mat_t Amod) noexcept
    void fmpz_mat_CRT_ui(fmpz_mat_t res, const fmpz_mat_t mat1, const fmpz_t m1, const nmod_mat_t mat2, int sign) noexcept
    void fmpz_mat_multi_mod_ui_precomp(nmod_mat_t * residues, slong nres, const fmpz_mat_t mat, const fmpz_comb_t comb, fmpz_comb_temp_t temp) noexcept
    void fmpz_mat_multi_mod_ui(nmod_mat_t * residues, slong nres, const fmpz_mat_t mat) noexcept
    void fmpz_mat_multi_CRT_ui_precomp(fmpz_mat_t mat, nmod_mat_t * const residues, slong nres, const fmpz_comb_t comb, fmpz_comb_temp_t temp, int sign) noexcept
    void fmpz_mat_multi_CRT_ui(fmpz_mat_t mat, nmod_mat_t * const residues, slong nres, int sign) noexcept
    void fmpz_mat_add(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void fmpz_mat_sub(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void fmpz_mat_neg(fmpz_mat_t B, const fmpz_mat_t A) noexcept
    void fmpz_mat_scalar_mul_si(fmpz_mat_t B, const fmpz_mat_t A, slong c) noexcept
    void fmpz_mat_scalar_mul_ui(fmpz_mat_t B, const fmpz_mat_t A, ulong c) noexcept
    void fmpz_mat_scalar_mul_fmpz(fmpz_mat_t B, const fmpz_mat_t A, const fmpz_t c) noexcept
    void fmpz_mat_scalar_addmul_si(fmpz_mat_t B, const fmpz_mat_t A, slong c) noexcept
    void fmpz_mat_scalar_addmul_ui(fmpz_mat_t B, const fmpz_mat_t A, ulong c) noexcept
    void fmpz_mat_scalar_addmul_fmpz(fmpz_mat_t B, const fmpz_mat_t A, const fmpz_t c) noexcept
    void fmpz_mat_scalar_submul_si(fmpz_mat_t B, const fmpz_mat_t A, slong c) noexcept
    void fmpz_mat_scalar_submul_ui(fmpz_mat_t B, const fmpz_mat_t A, ulong c) noexcept
    void fmpz_mat_scalar_submul_fmpz(fmpz_mat_t B, const fmpz_mat_t A, const fmpz_t c) noexcept
    void fmpz_mat_scalar_addmul_nmod_mat_ui(fmpz_mat_t B, const nmod_mat_t A, ulong c) noexcept
    void fmpz_mat_scalar_addmul_nmod_mat_fmpz(fmpz_mat_t B, const nmod_mat_t A, const fmpz_t c) noexcept
    void fmpz_mat_scalar_divexact_si(fmpz_mat_t B, const fmpz_mat_t A, slong c) noexcept
    void fmpz_mat_scalar_divexact_ui(fmpz_mat_t B, const fmpz_mat_t A, ulong c) noexcept
    void fmpz_mat_scalar_divexact_fmpz(fmpz_mat_t B, const fmpz_mat_t A, const fmpz_t c) noexcept
    void fmpz_mat_scalar_mul_2exp(fmpz_mat_t B, const fmpz_mat_t A, ulong exp) noexcept
    void fmpz_mat_scalar_tdiv_q_2exp(fmpz_mat_t B, const fmpz_mat_t A, ulong exp) noexcept
    void fmpz_mat_scalar_smod(fmpz_mat_t B, const fmpz_mat_t A, const fmpz_t P) noexcept
    void fmpz_mat_mul(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void fmpz_mat_mul_classical(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void fmpz_mat_mul_strassen(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void _fmpz_mat_mul_multi_mod(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B, int sign, flint_bitcnt_t bits) noexcept
    void fmpz_mat_mul_multi_mod(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    int fmpz_mat_mul_blas(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void fmpz_mat_mul_fft(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void fmpz_mat_sqr(fmpz_mat_t B, const fmpz_mat_t A) noexcept
    void fmpz_mat_sqr_bodrato(fmpz_mat_t B, const fmpz_mat_t A) noexcept
    void fmpz_mat_pow(fmpz_mat_t B, const fmpz_mat_t A, ulong e) noexcept
    void _fmpz_mat_mul_small(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void _fmpz_mat_mul_double_word(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void fmpz_mat_mul_fmpz_vec(fmpz * c, const fmpz_mat_t A, const fmpz * b, slong blen) noexcept
    void fmpz_mat_mul_fmpz_vec_ptr(fmpz * const * c, const fmpz_mat_t A, const fmpz * const * b, slong blen) noexcept
    void fmpz_mat_fmpz_vec_mul(fmpz * c, const fmpz * a, slong alen, const fmpz_mat_t B) noexcept
    void fmpz_mat_fmpz_vec_mul_ptr(fmpz * const * c, const fmpz * const * a, slong alen, const fmpz_mat_t B) noexcept
    int fmpz_mat_inv(fmpz_mat_t Ainv, fmpz_t den, const fmpz_mat_t A) noexcept
    void fmpz_mat_kronecker_product(fmpz_mat_t C, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void fmpz_mat_content(fmpz_t mat_gcd, const fmpz_mat_t A) noexcept
    void fmpz_mat_trace(fmpz_t trace, const fmpz_mat_t mat) noexcept
    void fmpz_mat_det(fmpz_t det, const fmpz_mat_t A) noexcept
    void fmpz_mat_det_cofactor(fmpz_t det, const fmpz_mat_t A) noexcept
    void fmpz_mat_det_bareiss(fmpz_t det, const fmpz_mat_t A) noexcept
    void fmpz_mat_det_modular(fmpz_t det, const fmpz_mat_t A, int proved) noexcept
    void fmpz_mat_det_modular_accelerated(fmpz_t det, const fmpz_mat_t A, int proved) noexcept
    void fmpz_mat_det_modular_given_divisor(fmpz_t det, const fmpz_mat_t A, const fmpz_t d, int proved) noexcept
    void fmpz_mat_det_bound(fmpz_t bound, const fmpz_mat_t A) noexcept
    void fmpz_mat_det_bound_nonzero(fmpz_t bound, const fmpz_mat_t A) noexcept
    void fmpz_mat_det_divisor(fmpz_t d, const fmpz_mat_t A) noexcept
    void fmpz_mat_similarity(fmpz_mat_t A, slong r, fmpz_t d) noexcept
    void _fmpz_mat_charpoly_berkowitz(fmpz * cp, const fmpz_mat_t mat) noexcept
    void fmpz_mat_charpoly_berkowitz(fmpz_poly_t cp, const fmpz_mat_t mat) noexcept
    void _fmpz_mat_charpoly_modular(fmpz * cp, const fmpz_mat_t mat) noexcept
    void fmpz_mat_charpoly_modular(fmpz_poly_t cp, const fmpz_mat_t mat) noexcept
    void _fmpz_mat_charpoly(fmpz * cp, const fmpz_mat_t mat) noexcept
    void fmpz_mat_charpoly(fmpz_poly_t cp, const fmpz_mat_t mat) noexcept
    slong _fmpz_mat_minpoly_modular(fmpz * cp, const fmpz_mat_t mat) noexcept
    void fmpz_mat_minpoly_modular(fmpz_poly_t cp, const fmpz_mat_t mat) noexcept
    slong _fmpz_mat_minpoly(fmpz * cp, const fmpz_mat_t mat) noexcept
    void fmpz_mat_minpoly(fmpz_poly_t cp, const fmpz_mat_t mat) noexcept
    slong fmpz_mat_rank(const fmpz_mat_t A) noexcept
    int fmpz_mat_col_partition(slong * part, fmpz_mat_t M, int short_circuit) noexcept
    int fmpz_mat_solve(fmpz_mat_t X, fmpz_t den, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    int fmpz_mat_solve_fflu(fmpz_mat_t X, fmpz_t den, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    int fmpz_mat_solve_fflu_precomp(fmpz_mat_t X, const slong * perm, const fmpz_mat_t FFLU, const fmpz_mat_t B) noexcept
    int fmpz_mat_solve_cramer(fmpz_mat_t X, fmpz_t den, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void fmpz_mat_solve_bound(fmpz_t N, fmpz_t D, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    int fmpz_mat_solve_dixon(fmpz_mat_t X, fmpz_t M, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    void _fmpz_mat_solve_dixon_den(fmpz_mat_t X, fmpz_t den, const fmpz_mat_t A, const fmpz_mat_t B, const nmod_mat_t Ainv, mp_limb_t p, const fmpz_t N, const fmpz_t D) noexcept
    int fmpz_mat_solve_dixon_den(fmpz_mat_t X, fmpz_t den, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    int fmpz_mat_solve_multi_mod_den(fmpz_mat_t X, fmpz_t den, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    int fmpz_mat_can_solve_multi_mod_den(fmpz_mat_t X, fmpz_t den, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    int fmpz_mat_can_solve_fflu(fmpz_mat_t X, fmpz_t den, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    int fmpz_mat_can_solve(fmpz_mat_t X, fmpz_t den, const fmpz_mat_t A, const fmpz_mat_t B) noexcept
    slong fmpz_mat_find_pivot_any(const fmpz_mat_t mat, slong start_row, slong end_row, slong c) noexcept
    slong fmpz_mat_fflu(fmpz_mat_t B, fmpz_t den, slong * perm, const fmpz_mat_t A, int rank_check) noexcept
    slong fmpz_mat_rref(fmpz_mat_t B, fmpz_t den, const fmpz_mat_t A) noexcept
    slong fmpz_mat_rref_fflu(fmpz_mat_t B, fmpz_t den, const fmpz_mat_t A) noexcept
    slong fmpz_mat_rref_mul(fmpz_mat_t B, fmpz_t den, const fmpz_mat_t A) noexcept
    bint fmpz_mat_is_in_rref_with_rank(const fmpz_mat_t A, const fmpz_t den, slong rank) noexcept
    slong fmpz_mat_rref_mod(slong * perm, fmpz_mat_t A, const fmpz_t p) noexcept
    void fmpz_mat_strong_echelon_form_mod(fmpz_mat_t A, const fmpz_t mod) noexcept
    slong fmpz_mat_howell_form_mod(fmpz_mat_t A, const fmpz_t mod) noexcept
    slong fmpz_mat_nullspace(fmpz_mat_t B, const fmpz_mat_t A) noexcept
    slong fmpz_mat_rref_fraction_free(slong * perm, fmpz_mat_t B, fmpz_t den, const fmpz_mat_t A) noexcept
    void fmpz_mat_hnf(fmpz_mat_t H, const fmpz_mat_t A) noexcept
    void fmpz_mat_hnf_transform(fmpz_mat_t H, fmpz_mat_t U, const fmpz_mat_t A) noexcept
    void fmpz_mat_hnf_classical(fmpz_mat_t H, const fmpz_mat_t A) noexcept
    void fmpz_mat_hnf_xgcd(fmpz_mat_t H, const fmpz_mat_t A) noexcept
    void fmpz_mat_hnf_modular(fmpz_mat_t H, const fmpz_mat_t A, const fmpz_t D) noexcept
    void fmpz_mat_hnf_modular_eldiv(fmpz_mat_t A, const fmpz_t D) noexcept
    void fmpz_mat_hnf_minors(fmpz_mat_t H, const fmpz_mat_t A) noexcept
    void fmpz_mat_hnf_pernet_stein(fmpz_mat_t H, const fmpz_mat_t A, flint_rand_t state) noexcept
    bint fmpz_mat_is_in_hnf(const fmpz_mat_t A) noexcept
    void fmpz_mat_snf(fmpz_mat_t S, const fmpz_mat_t A) noexcept
    void fmpz_mat_snf_diagonal(fmpz_mat_t S, const fmpz_mat_t A) noexcept
    void fmpz_mat_snf_kannan_bachem(fmpz_mat_t S, const fmpz_mat_t A) noexcept
    void fmpz_mat_snf_iliopoulos(fmpz_mat_t S, const fmpz_mat_t A, const fmpz_t mod) noexcept
    bint fmpz_mat_is_in_snf(const fmpz_mat_t A) noexcept
    void fmpz_mat_gram(fmpz_mat_t B, const fmpz_mat_t A) noexcept
    bint fmpz_mat_is_hadamard(const fmpz_mat_t H) noexcept
    int fmpz_mat_hadamard(fmpz_mat_t H) noexcept
    int fmpz_mat_get_d_mat(d_mat_t B, const fmpz_mat_t A) noexcept
    int fmpz_mat_get_d_mat_transpose(d_mat_t B, const fmpz_mat_t A) noexcept
    void fmpz_mat_is_spd(const fmpz_mat_t A) noexcept
    void fmpz_mat_chol_d(d_mat_t R, const fmpz_mat_t A) noexcept
    bint fmpz_mat_is_reduced(const fmpz_mat_t A, double delta, double eta) noexcept
    bint fmpz_mat_is_reduced_gram(const fmpz_mat_t A, double delta, double eta) noexcept
    bint fmpz_mat_is_reduced_with_removal(const fmpz_mat_t A, double delta, double eta, const fmpz_t gs_B, int newd) noexcept
    bint fmpz_mat_is_reduced_gram_with_removal(const fmpz_mat_t A, double delta, double eta, const fmpz_t gs_B, int newd) noexcept
    void fmpz_mat_lll_original(fmpz_mat_t A, const fmpq_t delta, const fmpq_t eta) noexcept
    void fmpz_mat_lll_storjohann(fmpz_mat_t A, const fmpq_t delta, const fmpq_t eta) noexcept

from .fmpz_mat_macros cimport *



# sage_setup: distribution = sagemath-flint
r"""
Compute Hilbert series of monomial ideals

This implementation was provided at :issue:`26243` and is supposed to be a way
out when Singular fails with an int overflow, which will regularly be the case
in any example with more than 34 variables.
"""
#*****************************************************************************
#
#    Tools to compute Hilbert Poincaré series of monomial ideals
#
#    Copyright (C) 2018 Simon A. King <simon.king@uni-jena.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#
#*****************************************************************************

import sage.interfaces.abc

from sage.rings.polynomial.polydict cimport ETuple
from sage.rings.polynomial.polynomial_integer_dense_flint cimport Polynomial_integer_dense_flint

from cysignals.memory cimport sig_malloc
from cpython.list cimport PyList_GET_ITEM

# A boilerplate class to implement a binary tree
# with some additional information

from sage.libs.flint.fmpz_poly cimport *
from sage.libs.flint.types cimport fmpz_poly_t

cdef class Node:
    """
    A node of a binary tree

    It has slots for data that allow to recursively compute
    the first Hilbert series of a monomial ideal.
    """
    cdef Node Back, Left, Right
    cdef list Id
    cdef fmpz_poly_t LMult
    cdef fmpz_poly_t RMult
    cdef fmpz_poly_t LeftFHS

    def __cinit__(self):
        fmpz_poly_init(self.LMult)
        fmpz_poly_init(self.RMult)
        fmpz_poly_init(self.LeftFHS)

    def __dealloc__(self):
        fmpz_poly_clear(self.LMult)
        fmpz_poly_clear(self.RMult)
        fmpz_poly_clear(self.LeftFHS)

cdef inline size_t median(list v) noexcept:
    """
    Specialized version of :func:`from sage.stats.basic_stats.median`.
    """
    cdef list values = sorted(v)
    cdef size_t nvals = len(values)
    if nvals % 2 == 1:
        return values[(nvals+1)//2 - 1]
    else:
        lower = values[(nvals+1)//2 - 1]
        upper = values[nvals // 2]
        return (lower + upper) // 2

###
#   cdef functions related with lists of monomials
###

cdef inline bint indivisible_in_list(ETuple m, list L, size_t i) noexcept:
    """
    Return if ``m`` divisible by any monomial in ``L[:i]``.
    """
    cdef size_t j
    for j in range(i):
        if (<ETuple>PyList_GET_ITEM(L,j)).divides(m):
            return False
    return True

cdef inline list interred(list L):
    """
    Return interreduction of a list of monomials.

    .. NOTE::

        The given list will be sorted in-place.

    INPUT:

    - ``L`` -- list of :class:`~sage.rings.polynomial.polydict.ETuple`

    OUTPUT:

    The interreduced list, where we interprete each ETuple as
    a monomial in a multivariate ring.
    """
    # First, we sort L ascendingly by total unweighted degree.
    # Afterwards, no monomial in L is divisible by a monomial
    # that appears later in L.
    if not L:
        return []
    L.sort(key=ETuple._unweighted_degree)
    cdef size_t i
    cdef ETuple m
    cdef list result = [<ETuple> PyList_GET_ITEM(L, 0)]
    for i in range(1, len(L)):
        m = <ETuple> PyList_GET_ITEM(L, i)
        if indivisible_in_list(m, L, i):
            result.append(m)
    return result

cdef list quotient(list L, ETuple m):
    """
    Return the quotient of the ideal represented by ``L`` and the
    monomial represented by ``m``.
    """
    cdef list result = L[:len(L)] # creates a copy
    cdef size_t i
    for i in range(len(L)):
        result.append((<ETuple>PyList_GET_ITEM(L,i)).divide_by_gcd(m))
    return interred(result)

cdef list quotient_by_var(list L, size_t index):
    """
    Return the quotient of the ideal represented by ``L`` and the
    variable number ``index``.
    """
    cdef list result = L[:len(L)] # creates a copy
    cdef size_t i
    for i in range(len(L)):
        if (<ETuple> PyList_GET_ITEM(L, i)).get_exp(index):
            result.append((<ETuple> PyList_GET_ITEM(L, i)).divide_by_var(index))
    return interred(result)

cdef ETuple sum_from_list(list L, size_t s, size_t l):
    """
    Compute the vector sum of the ETuples in ``L[s:s+l]`` in a balanced way.
    """
    if l == 1:
        return <ETuple>PyList_GET_ITEM(L,s)
    if l == 2:
        return (<ETuple>PyList_GET_ITEM(L,s)).eadd(<ETuple>PyList_GET_ITEM(L,s+1))
    cdef size_t l2 = l // 2
    cdef ETuple m1, m2
    m1 = sum_from_list(L, s, l2)
    m2 = sum_from_list(L, s+l2, l-l2)
    return m1.eadd(m2)

cdef bint HilbertBaseCase(Polynomial_integer_dense_flint fhs, Node D, tuple w) noexcept:
    """
    Try to compute the first Hilbert series of ``D.Id``, or return
    ``NotImplemented``.

    The third parameter ``w`` is a tuple of integers, the degree weights
    to be used.

    ``fhs`` is supposed to be zero. In some base cases, ``fhs`` will be
    changed to the value of the Hilbert series and ``True`` is returned.
    Otherwiese, ``False`` is returned.
    """
    cdef size_t i, j, exp
    # First, the easiest cases:
    if not D.Id: # The zero ideal
        fmpz_poly_set_coeff_si(fhs._poly, 0, 1) # = PR(1)
        return True
    cdef ETuple m = <ETuple>PyList_GET_ITEM(D.Id, len(D.Id)-1)
    if m._nonzero == 0: # The one ideal
        return True

    # Second, another reasy case: D.Id is generated by variables.
    # D.Id is sorted ascendingly. Hence, if the last generator is a single
    # variable, then ALL are.
    cdef fmpz_poly_t poly_tmp
    if m._nonzero == 1 and m._data[1] == 1:
        fmpz_poly_init(poly_tmp)
        fmpz_poly_set_coeff_si(poly_tmp, 0, 1)
        fmpz_poly_set_coeff_si(fhs._poly, 0, 1) # = PR(1)
        if w is None:
            for i in range(len(D.Id)):
                exp = (<ETuple>PyList_GET_ITEM(D.Id, i)).unweighted_degree()
                fmpz_poly_set_coeff_si(poly_tmp, exp, -1)
                fmpz_poly_mul(fhs._poly, fhs._poly, poly_tmp)
                fmpz_poly_set_coeff_si(poly_tmp, exp, 0)
        else:
            for i in range(len(D.Id)):
                exp = (<ETuple>PyList_GET_ITEM(D.Id, i)).weighted_degree(w)
                fmpz_poly_set_coeff_si(poly_tmp, exp, -1)
                fmpz_poly_mul(fhs._poly, fhs._poly, poly_tmp)
                fmpz_poly_set_coeff_si(poly_tmp, exp, 0)
        fmpz_poly_clear(poly_tmp)
        return True # PR.prod([(1-t**degree(m,w)) for m in D.Id])

    # Thirdly, we test for proper powers of single variables.
    cdef bint easy = True
    for i in range(len(D.Id)):
        m = <ETuple>PyList_GET_ITEM(D.Id, i) # will be used later when we are breaking
        if m._nonzero > 1: # i.e., the generator contains more than a single var
            easy = False
            break
    if easy:
        # The ideal is generated by some powers of single variables, i.e., it splits.
        fmpz_poly_init(poly_tmp)
        fmpz_poly_set_coeff_si(poly_tmp, 0, 1)
        fmpz_poly_set_coeff_si(fhs._poly, 0, 1) # = PR(1)
        if w is None:
            for i in range(len(D.Id)):
                exp = (<ETuple>PyList_GET_ITEM(D.Id, i)).unweighted_degree()
                fmpz_poly_set_coeff_si(poly_tmp, exp, -1)
                fmpz_poly_mul(fhs._poly, fhs._poly, poly_tmp)
                fmpz_poly_set_coeff_si(poly_tmp, exp, 0)
        else:
            for i in range(len(D.Id)):
                exp = (<ETuple>PyList_GET_ITEM(D.Id, i)).weighted_degree(w)
                fmpz_poly_set_coeff_si(poly_tmp, exp, -1)
                fmpz_poly_mul(fhs._poly, fhs._poly, poly_tmp)
                fmpz_poly_set_coeff_si(poly_tmp, exp, 0)
        fmpz_poly_clear(poly_tmp)
        return True # PR.prod([(1-t**degree(m,w)) for m in D.Id])

    easy = True
    cdef ETuple m2
    for j in range(i+1, len(D.Id)):
        if (<ETuple>PyList_GET_ITEM(D.Id,j))._nonzero > 1:
            # i.e., another generator contains more than a single var
            easy = False
            break
    cdef fmpz_poly_t FirstSummand, SecondSummand
    if easy:
        # The ideal only has a single non-simple power, in position i.
        # We have preserved that monomial, it is called m.
        fmpz_poly_init(poly_tmp)
        fmpz_poly_init(FirstSummand)
        fmpz_poly_init(SecondSummand)
        fmpz_poly_set_coeff_si(poly_tmp, 0, 1)
        fmpz_poly_set_coeff_si(FirstSummand, 0, 1)
        if w is None:
            fmpz_poly_set_coeff_si(SecondSummand, m.unweighted_degree(), -1)
            # Since the ideal is interreduced and all other monomials are
            # simple powers, we have the following formula:
            #~ return prod([(1-t**degree(m2,w)) for m2 in D.Id if m2 is not m])
            #~   - t**degree(m,w)*prod((1-t**quotient_degree(m2,m,w)) for m2 in D.Id if m is not m2)
            for j in range(len(D.Id)):
                if i != j:
                    m2 = <ETuple>PyList_GET_ITEM(D.Id,j)
                    exp = m2.unweighted_degree()
                    fmpz_poly_set_coeff_si(poly_tmp, exp, -1)
                    fmpz_poly_mul(FirstSummand, FirstSummand, poly_tmp)
                    fmpz_poly_set_coeff_si(poly_tmp, exp, 0)
                    exp = m2.unweighted_quotient_degree(m)
                    fmpz_poly_set_coeff_si(poly_tmp, exp, -1)
                    fmpz_poly_mul(SecondSummand, SecondSummand, poly_tmp)
                    fmpz_poly_set_coeff_si(poly_tmp, exp, 0)
        else:
            fmpz_poly_set_coeff_si(SecondSummand, m.weighted_degree(w), -1)
            # Since the ideal is interreduced and all other monomials are
            # simple powers, we have the following formula:
            #~ return prod([(1-t**degree(m2,w)) for m2 in D.Id if m2 is not m])
            #~   - t**degree(m,w)*prod((1-t**quotient_degree(m2,m,w)) for m2 in D.Id if m is not m2)
            for j in range(len(D.Id)):
                if i != j:
                    m2 = <ETuple>PyList_GET_ITEM(D.Id,j)
                    exp = m2.weighted_degree(w)
                    fmpz_poly_set_coeff_si(poly_tmp, exp, -1)
                    fmpz_poly_mul(FirstSummand, FirstSummand, poly_tmp)
                    fmpz_poly_set_coeff_si(poly_tmp, exp, 0)
                    exp = m2.weighted_quotient_degree(m,w)
                    fmpz_poly_set_coeff_si(poly_tmp, exp, -1)
                    fmpz_poly_mul(SecondSummand, SecondSummand, poly_tmp)
                    fmpz_poly_set_coeff_si(poly_tmp, exp, 0)
        fmpz_poly_clear(poly_tmp)
        fmpz_poly_add(fhs._poly, fhs._poly, FirstSummand)
        fmpz_poly_add(fhs._poly, fhs._poly, SecondSummand)
        fmpz_poly_clear(FirstSummand)
        fmpz_poly_clear(SecondSummand)
        return True
    # We are in a truly difficult case and give up for now...
    return False

cdef make_children(Node D, tuple w):
    """
    Create child nodes in ``D`` that allow to compute the first Hilbert
    series of ``D.Id``.

    Basically, the first Hilbert series of ``D.Id`` will be
    ``D.LMult`` times the first Hilbert series of ``D.Left.Id``,
    possibly plus ``D.RMult`` times the first Hilbert series of ``D.Right.Id``
    if ``D.Right`` is not ``None``.
    """
    cdef size_t j, m
    cdef int i
    # Determine the variable that appears most often in the monomials.
    # If "most often" means "only once", then instead we choose a variable that is
    # guaranteed to appear in a composed monomial.
    # We will raise it to a reasonably high power that still guarantees that
    # many monomials will be divisible by it.
    cdef ETuple all_exponents = sum_from_list(D.Id, 0, len(D.Id))
    m = 0
    cdef list max_exponents = []
    for i in range(0, 2*all_exponents._nonzero, 2):
        j = all_exponents._data[i+1]
        if j > m:
            max_exponents = [all_exponents._data[i]]
            m = j
        elif j == m:
            max_exponents.append(all_exponents._data[i])
    cdef size_t e # will be the exponent, if the monomial used for cutting is power of a variable
    cdef ETuple cut
    cdef list Id2
    cdef size_t cut_deg
    # Cases:
    # - m==1, which means that all variables occur at most once.
    #   => we cut by a variable that appears in a decomposable generator
    # - max_exponents = [j]
    #   => cut = var(j)**e, where e is the median of all j-exponents
    # - max_exponents = [j1,...,jk]
    #   => cut = prod([var(j1),...,var(jk)]) or something of that type.
    if m == 1:
        # D.Id is sorted, which means that the last generator is decomposable
        all_exponents = <ETuple> PyList_GET_ITEM(D.Id, len(D.Id)-1)
        j = all_exponents._data[2*all_exponents._nonzero-2]
        cut = all_exponents._new()
        cut._nonzero = 1
        cut._data = <int*> sig_malloc(sizeof(int)*2)
        cut._data[0] = j
        cut._data[1] = 1
        if w is None:
            cut_deg = cut.unweighted_degree()
        else:
            cut_deg = cut.weighted_degree(w)
        # var(j) *only* appears in D.Id[-1]. Hence, D.Id+var(j) will be a split case,
        # with var(j) and D.Id[:-1]. So, we do the splitting right now.
        # Only the last generator contains var(j). Hence, D.Id/var(j) is obtained
        # from D.Id by adding the quotient of its last generator divided by var(j),
        # of course followed by interreduction.
        #~ D.LMult = 1 - t**degree(cut,w)
        fmpz_poly_set_coeff_si(D.LMult, 0, 1)
        fmpz_poly_set_coeff_si(D.LMult, cut_deg, -1)
        D.Left  = Node.__new__(Node)
        D.Left.Id = D.Id[:len(D.Id)-1]
        D.Left.Back = D
        Id2 = D.Id[:len(D.Id)-1]
        Id2.append(all_exponents.divide_by_var(j))
        D.Right = Node.__new__(Node)
        D.Right.Id = interred(Id2)
        D.Right.Back = D
        #~ D.RMult = 1 - D.LMult
        fmpz_poly_set_coeff_si(D.RMult, cut_deg, 1)
    else:
        j = max_exponents[0]
        e = median([(<ETuple> mon).get_exp(j) for mon in D.Id if (<ETuple> mon).get_exp(j)])
        cut = all_exponents._new()
        cut._nonzero = 1
        cut._data = <int*> sig_malloc(sizeof(int)*2)
        cut._data[0] = j
        cut._data[1] = e
        if w is None:
            cut_deg = cut.unweighted_degree()
        else:
            cut_deg = cut.weighted_degree(w)
        try:
            i = D.Id.index(cut)
        except ValueError:
            i = -1
        if i >= 0:
            # var(j)**e is a generator. Hence, e is the maximal exponent of
            # var(j) in D.Id, by D.Id being interreduced. But it also is the
            # truncated median, hence, there cannot be smaller exponents
            # (for otherwise the median would be strictly smaller than the
            # maximum). Conclusion: var(j) only appears in the generator var(j)**e
            #  -- we have a split case.
            Id2 = D.Id[:len(D.Id)]
            Id2.pop(i)
            #~ D.LMult = 1-t**degree(cut,w)
            fmpz_poly_set_coeff_si(D.LMult, 0, 1)
            fmpz_poly_set_coeff_si(D.LMult, cut_deg, -1)
            D.Left  = Node.__new__(Node)
            D.Left.Id = Id2
            D.Left.Back = D
            D.Right = None
        else:
            cut = all_exponents._new()
            cut._nonzero = 1
            cut._data = <int*> sig_malloc(sizeof(int)*2)
            cut._data[0] = j
            cut._data[1] = e
            if e > 1:
                # D.LMult = 1
                fmpz_poly_set_coeff_si(D.LMult, 0, 1)
                Id2 = D.Id[:len(D.Id)]
                Id2.append(cut)
                D.Left  = Node.__new__(Node)
                D.Left.Id = interred(Id2)
                D.Left.Back = D
                D.Right = Node.__new__(Node)
                D.Right.Id = quotient(D.Id,cut)
                D.Right.Back = D
            else:
                # m > 1, therefore var(j) cannot be a generator (D.Id is interreduced).
                # D.Id+var(j) will be a split case. So, we do the splitting right now.
                #~ D.LMult = 1-t**(1 if w is None else w[j])
                fmpz_poly_set_coeff_si(D.LMult, 0, 1)
                fmpz_poly_set_coeff_si(D.LMult, (1 if w is None else w[j]), -1)
                D.Left  = Node.__new__(Node)
                D.Left.Id = [(<ETuple> mon) for mon in D.Id if (<ETuple> mon).get_exp(j) == 0]
                D.Left.Back = D
                D.Right = Node.__new__(Node)
                D.Right.Id = quotient_by_var(D.Id, j)
                D.Right.Back = D
            #~ D.RMult = t**(e if w is None else e*w[j])
            fmpz_poly_set_coeff_si(D.RMult, (e if w is None else e * w[j]), 1)
     #else:
     #    It may be a good idea to form the product of some of the most
     #    frequent variables. But this isn't implemented yet. TODO?


def first_hilbert_series(I, grading=None, return_grading=False):
    """
    Return the first Hilbert series of the given monomial ideal.

    INPUT:

    - ``I`` -- a monomial ideal (possibly defined in singular)
    - ``grading`` -- (optional) a list or tuple of integers used as
      degree weights
    - ``return_grading`` -- boolean (default: ``False``); whether to return the
      grading

    OUTPUT:

    A univariate polynomial, namely the first Hilbert function of ``I``, and
    if ``return_grading==True`` also the grading used to compute the series.

    EXAMPLES::

        sage: from sage.rings.polynomial.hilbert import first_hilbert_series

        sage: # needs sage.libs.singular
        sage: R = singular.ring(0,'(x,y,z)','dp')
        sage: I = singular.ideal(['x^2','y^2','z^2'])
        sage: first_hilbert_series(I)
        -t^6 + 3*t^4 - 3*t^2 + 1
        sage: first_hilbert_series(I, return_grading=True)
        (-t^6 + 3*t^4 - 3*t^2 + 1, (1, 1, 1))
        sage: first_hilbert_series(I, grading=(1,2,3))
        -t^12 + t^10 + t^8 - t^4 - t^2 + 1

    TESTS:

    We test against some corner cases::

        sage: R.<x,y,z> = PolynomialRing(QQ)
        sage: I = 0*R
        sage: first_hilbert_series(I)
        1
        sage: first_hilbert_series(singular(I))                                         # needs sage.libs.singular
        1
        sage: I = 1*R
        sage: first_hilbert_series(I)
        0
        sage: first_hilbert_series(singular(I))                                         # needs sage.libs.singular
        0
    """
    from sage.rings.integer_ring import ZZ
    from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
    PR = PolynomialRing(ZZ, 't')
    cdef Node AN
    # The "active node". If a recursive computation is needed, it will be equipped
    # with a 'Left' and a 'Right' child node, and some 'Multipliers'. Later, the first Hilbert
    # series of the left child node will be stored in 'LeftFHS', and together with
    # the first Hilbert series of the right child node and the multiplier yields
    # the first Hilbert series of 'Id'.
    cdef tuple w
    cdef Polynomial_integer_dense_flint fhs = Polynomial_integer_dense_flint.__new__(Polynomial_integer_dense_flint)
    fhs._parent = PR
    fhs._is_gen = 0
    if isinstance(I, sage.interfaces.abc.SingularElement):
        S = I._check_valid()
        # First, we need to deal with quotient rings, which also covers the case
        # of graded commutative rings that arise as cohomology rings in odd characteristic.
        # We replace everything by a commutative version of the quotient ring.
        br = S('basering')
        if S.eval('isQuotientRing(basering)')=='1':
            L = S('ringlist(basering)')
            R = S('ring(list(%s[1..3],ideal(0)))' % L.name())
            R.set_ring()
            I = S('fetch(%s,%s)+ideal(%s)' % (br.name(), I.name(), br.name()))

        I = [ETuple([int(x) for x in S.eval('string(leadexp({}[{}]))'.format(I.name(), i)).split(',')])
              for i in range(1,int(S.eval('size({})'.format(I.name())))+1)]
        br.set_ring()
        if grading is None:
            w = tuple(int(S.eval('deg(var({}))'.format(i))) for i in range(1,int(S.eval('nvars(basering)'))+1))
        else:
            w = tuple(grading)
    else:
        try:
            I = [bla.exponents()[0] for bla in I if bla]
        except TypeError:
            I = [bla.exponents()[0] for bla in I.gens() if bla]
        if grading is not None:
            w = tuple(grading)
        else:
            w = None

    AN = Node.__new__(Node)
    AN.Id = interred(I)
    AN.Back = None

    # Invariant of this function:
    # At each point, got_result will be false, or got_result is true and fhs
    # is the first Hilbert series of AN.
    cdef bint got_result = HilbertBaseCase(fhs, AN, w)
    while True:
        if not got_result:
            make_children(AN, w)
            AN = AN.Left
            got_result = HilbertBaseCase(fhs, AN, w)
        else:
            if AN.Back is None: # We are back on top, i.e., fhs is the First Hilber Series of I
                if return_grading:
                    return fhs, w
                else:
                    return fhs
            if AN is AN.Back.Left: # We store fhs and proceed to the sibling
                # ... unless there is no sibling
                if AN.Back.Right is None:
                    AN = AN.Back
                    #~ fhs *= AN.LMult
                    fmpz_poly_mul(fhs._poly, fhs._poly, AN.LMult)
                    got_result = True
                else:
                    fmpz_poly_set(AN.Back.LeftFHS, fhs._poly)
                    fmpz_poly_set_si(fhs._poly, 0)
                    AN = AN.Back.Right
                    AN.Back.Left = None
                    got_result = HilbertBaseCase(fhs, AN, w)
            else: # FHS of the left sibling is stored, of the right sibling is known.
                AN = AN.Back
                AN.Right = None
                #~ fhs = AN.LMult*AN.LeftFHS + AN.RMult*fhs
                fmpz_poly_mul(AN.LMult, AN.LMult, AN.LeftFHS)
                fmpz_poly_mul(AN.RMult, AN.RMult, fhs._poly)
                fmpz_poly_add(fhs._poly, AN.LMult, AN.RMult)
                got_result = True


def hilbert_poincare_series(I, grading=None):
    r"""
    Return the Hilbert Poincaré series of the given monomial ideal.

    INPUT:

    - ``I`` -- a monomial ideal (possibly defined in Singular)
    - ``grading`` -- (optional) a tuple of degree weights

    EXAMPLES::

        sage: # needs sage.libs.singular
        sage: from sage.rings.polynomial.hilbert import hilbert_poincare_series
        sage: R = PolynomialRing(QQ,'x',9)
        sage: I = [m.lm()
        ....:      for m in ((matrix(R, 3, R.gens())^2).list() * R).groebner_basis()] * R
        sage: hilbert_poincare_series(I)
        (t^7 - 3*t^6 + 2*t^5 + 2*t^4 - 2*t^3 + 6*t^2 + 5*t + 1)/(t^4 - 4*t^3 + 6*t^2 - 4*t + 1)
        sage: hilbert_poincare_series((R * R.gens())^2, grading=range(1,10))
        t^9 + t^8 + t^7 + t^6 + t^5 + t^4 + t^3 + t^2 + t + 1

    The following example is taken from :issue:`20145`::

        sage: # needs sage.libs.singular
        sage: n=4; m=11; P = PolynomialRing(QQ, n*m, "x"); x = P.gens(); M = Matrix(n, x)
        sage: from sage.rings.polynomial.hilbert import first_hilbert_series
        sage: I = P.ideal(M.minors(2))
        sage: J = P * [m.lm() for m in I.groebner_basis()]
        sage: hilbert_poincare_series(J).numerator()
        120*t^3 + 135*t^2 + 30*t + 1
        sage: hilbert_poincare_series(J).denominator().factor()
        (t - 1)^14
    """
    cdef Polynomial_integer_dense_flint HP
    HP, grading = first_hilbert_series(I, grading=grading, return_grading=True)
    PR = HP._parent

    # If grading was None, but the ideal lives in Singular, then grading is now
    # the degree vector of Singular's basering.
    # Otherwise, it may still be None.
    if grading is None:
        if HP.leading_coefficient() >= 0:
            return HP / ((1-PR.gen())**I.ring().ngens())
        return (-HP) / (-((1-PR.gen())**I.ring().ngens()))
    else:
        t = PR.gen()
        if HP.leading_coefficient() >= 0:
            return HP / PR.prod([(1-t**d) for d in grading])
        return (-HP) / (-PR.prod([(1-t**d) for d in grading]))



# sage_setup: distribution = sagemath-modules
r"""
Optimized counting of congruence solutions
"""
from sage.arith.misc import is_prime, kronecker as kronecker_symbol, valuation
from sage.rings.finite_rings.integer_mod import Mod
from sage.rings.finite_rings.integer_mod_ring import IntegerModRing


def count_modp__by_gauss_sum(n, p, m, Qdet):
    r"""
    Return the number of solutions of `Q(x) = m` over the finite field
    `\ZZ/p\ZZ`, where `p` is a prime number > 2 and `Q` is a non-degenerate
    quadratic form of dimension `n \geq 1` and has Gram determinant ``Qdet``.

    REFERENCE:

    These are defined in Table 1 on page 363 of [Hanke2004]_.

    INPUT:

    - ``n`` -- integer `\geq 1`
    - ``p`` -- a prime number > 2
    - ``m`` -- integer
    - ``Qdet`` -- a integer which is nonzero mod `p`

    OUTPUT: integer `\geq 0`

    EXAMPLES::

        sage: from sage.quadratic_forms.count_local_2 import count_modp__by_gauss_sum

        sage: count_modp__by_gauss_sum(3, 3, 0, 1)    # for Q = x^2 + y^2 + z^2  => Gram Det = 1 (mod 3)
        9
        sage: count_modp__by_gauss_sum(3, 3, 1, 1)    # for Q = x^2 + y^2 + z^2  => Gram Det = 1 (mod 3)
        6
        sage: count_modp__by_gauss_sum(3, 3, 2, 1)    # for Q = x^2 + y^2 + z^2  => Gram Det = 1 (mod 3)
        12

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: [Q.count_congruence_solutions(3, 1, m, None, None)
        ....:    == count_modp__by_gauss_sum(3, 3, m, 1)
        ....:  for m in range(3)]
        [True, True, True]


        sage: count_modp__by_gauss_sum(3, 3, 0, 2)    # for Q = x^2 + y^2 + 2*z^2  => Gram Det = 2 (mod 3)
        9
        sage: count_modp__by_gauss_sum(3, 3, 1, 2)    # for Q = x^2 + y^2 + 2*z^2  => Gram Det = 2 (mod 3)
        12
        sage: count_modp__by_gauss_sum(3, 3, 2, 2)    # for Q = x^2 + y^2 + 2*z^2  => Gram Det = 2 (mod 3)
        6

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,2])
        sage: [Q.count_congruence_solutions(3, 1, m, None, None)
        ....:    == count_modp__by_gauss_sum(3, 3, m, 2)
        ....:  for m in range(3)]
        [True, True, True]
    """
    # Check that Qdet is non-degenerate
    if Qdet % p == 0:
        raise RuntimeError("Qdet must be nonzero")

    # Check that p is prime > 2
    if not is_prime(p) or p == 2:
        raise RuntimeError("p must be a prime number > 2")

    # Check that n >= 1
    if n < 1:
        raise RuntimeError("the dimension n must be >= 1")

    # Compute the Gauss sum
    neg1 = -1
    if not m % p:
        if n % 2:
            count = p**(n - 1)
        else:
            count = p**(n - 1) + (p - 1) * (p**((n - 2) // 2)) * kronecker_symbol(((neg1**(n // 2)) * Qdet) % p, p)
    else:
        if n % 2:
            count = p**(n - 1) + p**((n - 1) // 2) * kronecker_symbol(((neg1**((n - 1) // 2)) * Qdet * m) % p, p)
        else:
            count = p**(n - 1) - p**((n - 2) // 2) * kronecker_symbol(((neg1**(n // 2)) * Qdet) % p, p)

    # Return the result
    return count


cdef CountAllLocalTypesNaive_cdef(Q, p, k, m, zvec, nzvec):
    """
    This Cython routine is documented in its Python wrapper method
    QuadraticForm.count_congruence_solutions_by_type().
    """
    cdef long n, i
    cdef long a, b    # Used to quickly evaluate Q(v)
    cdef long ptr     # Used to increment the vector
    cdef long solntype    # Used to store the kind of solution we find

    # Some shortcuts and definitions
    n = Q.dim()
    R = p ** k
    Q1 = Q.change_ring(IntegerModRing(R))

    # Initialize the counting vector
    count_vector = [0 for i in range(6)]

    # Initialize v = (0, ... , 0)
    v = [Mod(0, R) for i in range(n)]

    # Some declarations to speed up the loop
    R_n = R ** n
    m1 = Mod(m, R)

    # Count the local solutions
    for i in range(R_n):

        # Perform a carry (when value = R-1) until we can increment freely
        ptr = len(v)
        while ((ptr > 0) and (v[ptr - 1] == R - 1)):
            v[ptr - 1] += 1
            ptr += -1

        # Only increment if we're not already at the zero vector =)
        if ptr > 0:
            v[ptr - 1] += 1

        # Evaluate Q(v) quickly
        tmp_val = Mod(0, R)
        for a in range(n):
            for b in range(a, n):
                tmp_val += Q1[a, b] * v[a] * v[b]

        # Sort the solution by it's type
        # if Q1(v) == m1:
        if tmp_val == m1:
            solntype = local_solution_type_cdef(Q1, p, v, zvec, nzvec)
            if solntype != 0:
                count_vector[solntype] += 1

    # Generate the Bad-type and Total counts
    count_vector[3] = count_vector[4] + count_vector[5]
    count_vector[0] = count_vector[1] + count_vector[2] + count_vector[3]

    # Return the solution counts
    return count_vector


def CountAllLocalTypesNaive(Q, p, k, m, zvec, nzvec):
    r"""
    This is an internal routine, which is called by
    :meth:`sage.quadratic_forms.quadratic_form.QuadraticForm.count_congruence_solutions_by_type
    QuadraticForm.count_congruence_solutions_by_type`. See the documentation of
    that method for more details.

    INPUT:

    - ``Q`` -- quadratic form over `\ZZ`
    - ``p`` -- prime number > 0
    - ``k`` -- integer > 0
    - ``m`` -- integer (depending only on mod `p^k`)
    - ``zvec``, ``nzvec`` -- list of integers in ``range(Q.dim())``, or ``None``

    OUTPUT:

    a list of six integers `\ge 0` representing the solution types: ``[All,
    Good, Zero, Bad, BadI, BadII]``

    EXAMPLES::

        sage: from sage.quadratic_forms.count_local_2 import CountAllLocalTypesNaive
        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: CountAllLocalTypesNaive(Q, 3, 1, 1, None, None)
        [6, 6, 0, 0, 0, 0]
        sage: CountAllLocalTypesNaive(Q, 3, 1, 2, None, None)
        [6, 6, 0, 0, 0, 0]
        sage: CountAllLocalTypesNaive(Q, 3, 1, 0, None, None)
        [15, 12, 1, 2, 0, 2]
    """
    return CountAllLocalTypesNaive_cdef(Q, p, k, m, zvec, nzvec)


cdef local_solution_type_cdef(Q, p, w, zvec, nzvec):
    """
    Internal routine to check if a given solution vector `w` (of `Q(w) =
    m` mod `p^k`) is of a certain local type and satisfies certain
    congruence conditions mod `p`.

    NOTE: No internal checking is done to test if `p` is a prime >=2, or
    that Q has the same size as `w`.
    """
    cdef long i
    cdef long n

    n = Q.dim()

    # Check if the solution satisfies the zvec "zero" congruence conditions
    # (either zvec is empty or its components index the zero vector mod p)
    if zvec is None or not zvec:
        zero_flag = True
    else:
        zero_flag = False
        i = 0
        while i < len(zvec) and not w[zvec[i]] % p:  # Increment so long as our entry is zero (mod p)
            i += 1
        if i == len(zvec):      # If we make it through all entries then the solution is zero (mod p)
            zero_flag = True

    # DIAGNOSTIC
    # print("IsLocalSolutionType: Finished the Zero congruence condition test \n")

    if not zero_flag:
        return <long> 0

    # DIAGNOSTIC
    # print("IsLocalSolutionType: Passed the Zero congruence condition test \n")

    # Check if the solution satisfies the nzvec "nonzero" congruence conditions
    # (nzvec is non-empty and its components index a nonzero vector mod p)
    if nzvec is None:
        nonzero_flag = True
    elif len(nzvec) == 0:
        nonzero_flag = False           # Trivially no solutions in this case!
    else:
        nonzero_flag = False
        i = 0
        while not nonzero_flag and i < len(nzvec):
            if w[nzvec[i]] % p:
                nonzero_flag = True           # The nonzero condition is satisfied when we find one nonzero entry
            i += 1

    if not nonzero_flag:
        return <long> 0

    # Check if the solution has the appropriate (local) type:
    # -------------------------------------------------------

    # 1: Check Good-type
    for i in range(n):
        if w[i] % p and Q[i, i] % p:
            return <long> 1
    if p == 2:
        for i in range(n - 1):
            if Q[i, i + 1] % p and (w[i] % p or w[i + 1] % p):
                return <long> 1

    # 2: Check Zero-type
    Zero_flag = True
    for i in range(n):
        if w[i] % p:
            Zero_flag = False
    if Zero_flag:
        return <long> 2

    # Check if wS1 is zero or not
    wS1_nonzero_flag = False
    for i in range(n):

        # Compute the valuation of each index, allowing for off-diagonal terms
        if Q[i, i] == 0:
            if i == 0:
                val = valuation(Q[i, i + 1], p)   # Look at the term to the right
            elif i == n - 1:
                val = valuation(Q[i - 1, i], p)   # Look at the term above
            else:
                val = valuation(Q[i, i + 1] + Q[i - 1, i], p)  # Finds the valuation of the off-diagonal term since only one isn't zero
        else:
            val = valuation(Q[i, i], p)

        # Test each index
        if val == 1 and w[i] % p:
            wS1_nonzero_flag = True

    # 4: Check Bad-type I
    if wS1_nonzero_flag:
        return <long> 4

    # 5: Check Bad-type II
    if not wS1_nonzero_flag:
        return <long> 5

    # Error if we get here! =o
    print("   Solution vector is " + str(w))
    print("   and Q is \n" + str(Q) + "\n")
    raise RuntimeError("Error in IsLocalSolutionType: Should not execute this line... =( \n")


def count_all_local_good_types_normal_form(Q, p, k, m, zvec, nzvec):
    r"""
    This is an internal routine, which is called by
    :meth:`sage.quadratic_forms.quadratic_form.QuadraticForm.local_good_density_congruence_even
    QuadraticForm.local_good_density_congruence_even`. See the documentation of
    that method for more details.

    INPUT:

    - ``Q`` -- quadratic form over `\ZZ` in local normal form at p with no zero blocks mod `p^k`
    - ``p`` -- prime number > 0
    - ``k`` -- integer > 0
    - ``m`` -- non-negative integer (depending only on mod `p^k`)
    - ``zvec``, ``nzvec`` -- list of integers in ``range(Q.dim())``, or ``None``

    OUTPUT:

    a non-negative integer giving the number of solutions of Good type.

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: from sage.quadratic_forms.count_local_2 import count_all_local_good_types_normal_form
        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q_local_at2 = Q.local_normal_form(2)
        sage: Q_local_at3 = Q.local_normal_form(3)
        sage: count_all_local_good_types_normal_form(Q_local_at2, 2, 3, 3, None, None)
        64
        sage: count_all_local_good_types_normal_form(Q_local_at2, 2, 3, 3, [0], None)
        32
        sage: count_all_local_good_types_normal_form(Q_local_at3, 3, 2, 1, None, None)
        54
    """
    n = Q.dim()
    if n == 0:
        return 0

    m_range = p**k
    if zvec is None:
        zvec = []
    if nzvec is None:
        nzvec = []

    # determine local blocks
    blocks = []
    i = 0
    while i < n - 1:
        if Q[i, i + 1] != 0:
            blocks += [(i, i + 1)]
            i += 2
        else:
            blocks += [(i,)]
            i += 1
    if i < n:
        blocks += [(i,)]

    solutions = [[0, 0] for _ in range(m_range)]  # [good, not good]
    solutions[0][1] = 1
    for b in blocks:
        Q_part = Q.extract_variables(b)
        zvec_local = range(len(b)) if (b[0] in zvec) else None
        nzvec_local = range(len(b)) if (b[0] in nzvec) else None

        solutions_part = [[0, 0] for _ in range(m_range)]
        for m_part in range(m_range):
            cnt = CountAllLocalTypesNaive(Q_part, p, k, m_part, zvec_local, nzvec_local)
            solutions_part[m_part][0] = cnt[1]
            solutions_part[m_part][1] = cnt[0] - cnt[1]

        # compute convolution of counts
        solutions_new = [[0, 0] for _ in range(m_range)]
        for m1 in range(m_range):
            for m2 in range(m_range):
                total = (solutions[m1][0] + solutions[m1][1]) * (solutions_part[m2][0] + solutions_part[m2][1])
                good = total - solutions[m1][1] * solutions_part[m2][1]
                solutions_new[(m1 + m2) % m_range][0] += good
                solutions_new[(m1 + m2) % m_range][1] += total - good
        solutions = solutions_new
    return solutions[m % m_range][0]



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.groups
r"""
Invariant modules
"""

# ****************************************************************************
#       Copyright (C) 2021 Trevor K. Karn <karnx018 at umn.edu>
#                     2021 Travis Scrimshaw
#                     2022 Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import operator
from sage.modules.with_basis.subquotient import SubmoduleWithBasis
from sage.modules.with_basis.representation import Representation
from sage.categories.finitely_generated_semigroups import FinitelyGeneratedSemigroups
from sage.categories.finite_dimensional_modules_with_basis import FiniteDimensionalModulesWithBasis
from sage.sets.family import Family
from sage.matrix.constructor import Matrix


class FiniteDimensionalInvariantModule(SubmoduleWithBasis):
    r"""
    The invariant submodule under a semigroup action.

    When a semigroup `S` acts on a module `M`, the invariant module is the
    set of elements `m \in M` such that `s \cdot m = m` for all `s \in S`:

    .. MATH::

        M^S := \{m \in M : s \cdot m = m,\, \forall s \in S \}.

    INPUT:

    - ``M`` -- a module in the category of
      :class:`~sage.categories.finite_dimensional_modules_with_basis.FiniteDimensionalModulesWithBasis`

    - ``S`` -- a semigroup in the category of
      :class:`~sage.categories.finitely_generated_semigroups.FinitelyGeneratedSemigroups`

    - ``action`` -- (default: ``operator.mul``) the action of ``S`` on ``M``

    - ``side`` -- (default: ``'left'``) the side on which ``S`` acts

    EXAMPLES:

    First, we create the invariant defined by the cyclic group action on the
    free module with basis `\{1,2,3\}`::

        sage: G = CyclicPermutationGroup(3)
        sage: M = CombinatorialFreeModule(QQ, [1,2,3], prefix='M')
        sage: action = lambda g, m: M.monomial(g(m))  # cyclically permute coordinates

    In order to give the module an action of ``G``, we create a
    :class:`~sage.modules.with_basis.representation.Representation`::

        sage: from sage.modules.with_basis.representation import Representation
        sage: R = Representation(G, M, action)
        sage: I = R.invariant_module()

    Then we can lift the basis from the invariant to the original module::

        sage: [I.lift(b) for b in I.basis()]
        [M[1] + M[2] + M[3]]

    The we could also have the action be a right-action, instead of the
    default left-action::

        sage: def rt_action(g, m): return M.monomial(g(m))  # cyclically permute coordinates
        sage: R = Representation(G, M, rt_action, side='right')  # same as last but on right
        sage: g = G.an_element(); g
        (1,2,3)
        sage: r = R.an_element(); r
        2*M[1] + 2*M[2] + 3*M[3]
        sage: R.side()
        'right'

    So now we can see that multiplication with ``g`` on the right sends
    ``M[1]`` to ``M[2]`` and so on::

        sage: r * g
        3*M[1] + 2*M[2] + 2*M[3]
        sage: I = R.invariant_module()
        sage: [I.lift(b) for b in I.basis()]
        [M[1] + M[2] + M[3]]

    Now we will take the regular representation of the symmetric group on
    three elements to be the module, and compute its invariant submodule::

        sage: # needs sage.combinat
        sage: G = SymmetricGroup(3)
        sage: R = G.regular_representation(QQ)
        sage: I = R.invariant_module()
        sage: [I.lift(b).to_vector() for b in I.basis()]
        [(1, 1, 1, 1, 1, 1)]

    We can also check the scalar multiplication by elements of the base ring
    (for this example, the rational field)::

        sage: [I.lift(3*b).to_vector() for b in I.basis()]                              # needs sage.combinat
        [(3, 3, 3, 3, 3, 3)]

    A more subtle example is the invariant submodule of a skew-commutative
    module, for example the exterior algebra `E[x_0,x_1,x_2]` generated
    by three elements::

        sage: G = CyclicPermutationGroup(3)
        sage: M = algebras.Exterior(QQ, 'x', 3)
        sage: def cyclic_ext_action(g, m):
        ....:     # cyclically permute generators
        ....:     return M.prod([M.monomial(FrozenBitset([g(j+1)-1])) for j in m])

    If you care about being able to exploit the algebra structure of the
    exterior algebra (i.e. if you want to multiply elements together), you
    should make sure the representation knows it is also an algebra with
    the semigroup action being by algebra endomorphisms::

        sage: cat = Algebras(QQ).WithBasis().FiniteDimensional()
        sage: R = Representation(G, M, cyclic_ext_action, category=cat)
        sage: I = R.invariant_module()

    We can express the basis in the ambient algebra (`E[x_0,x_1,x_2]`)::

        sage: [I.lift(b) for b in I.basis()]
        [1, x0 + x1 + x2, x0*x1 - x0*x2 + x1*x2, x0*x1*x2]

    or we can express the basis intrinsicallly to the invariant ``I``::

        sage: B = I.basis()
        sage: m = 3*B[0] + 2*B[1] + 7*B[3]

    This lifts to the exterior algebra::

        sage: I.lift(m)
        3 + 2*x0 + 7*x0*x1*x2 + 2*x1 + 2*x2

    We can also check using the invariant element ``m`` that arithmetic works::

        sage: m^2
        9*B[0] + 12*B[1] + 42*B[3]
        sage: m+m
        6*B[0] + 4*B[1] + 14*B[3]

    To see the actual elements expressed in the exterior algebra, we lift them
    again::

        sage: I.lift(m+m)
        6 + 4*x0 + 14*x0*x1*x2 + 4*x1 + 4*x2
        sage: 7*m
        21*B[0] + 14*B[1] + 49*B[3]
        sage: I.lift(7*m)
        21 + 14*x0 + 49*x0*x1*x2 + 14*x1 + 14*x2

    The classic example of an invariant module is the module of symmetric
    functions, which is the invariant module of polynomials whose variables
    are acted upon by permutation. We can create a module isomorphic to the
    homogeneous component of a a polynomial ring in `n` variable of a fixed
    degree `d` by looking at  weak  compositions of `d` of length `n`, which
    we consider as the exponent vector. For example, `x^2yz \in \QQ[x,y,z]`
    would have the exponent vector `(2,1,1)`. The vector `(2,1,1)` is a
    weak composition of `4`, with length `3`, and so we can think of it as
    being in the degree-`4` homogeneous component of a polynomial ring
    in three variables::

        sage: C = IntegerVectors(4, length=3, min_part=0)  # representing degree-4 monomials
        sage: M = CombinatorialFreeModule(QQ, C)  # isomorphic to deg-4 homog. polynomials
        sage: G = SymmetricGroup(3)
        sage: def perm_action(g, x): return M.monomial(C(g(list(x))))
        sage: perm_action(G((1,2,3)), C([4,3,2]))
        B[[3, 2, 4]]
        sage: R = Representation(G, M, perm_action)
        sage: I = R.invariant_module()
        sage: [I.lift(b) for b in I.basis()]
        [B[[0, 0, 4]] + B[[0, 4, 0]] + B[[4, 0, 0]],
         B[[0, 1, 3]] + B[[0, 3, 1]] + B[[1, 0, 3]]
         + B[[1, 3, 0]] + B[[3, 0, 1]] + B[[3, 1, 0]],
         B[[0, 2, 2]] + B[[2, 0, 2]] + B[[2, 2, 0]],
         B[[1, 1, 2]] + B[[1, 2, 1]] + B[[2, 1, 1]]]

    These are the monomial symmetric functions, which are a well-known
    basis for the symmetric functions. For comparison::

        sage: Sym = SymmetricFunctions(QQ)
        sage: m = Sym.monomial()
        sage: [m[mu].expand(3) for mu in Partitions(4)]
        [x0^4 + x1^4 + x2^4,
         x0^3*x1 + x0*x1^3 + x0^3*x2 + x1^3*x2 + x0*x2^3 + x1*x2^3,
         x0^2*x1^2 + x0^2*x2^2 + x1^2*x2^2,
         x0^2*x1*x2 + x0*x1^2*x2 + x0*x1*x2^2,
         0]

    .. NOTE::

        The current implementation works when `S` is a finitely-generated
        semigroup, and when `M` is a finite-dimensional free module with
        a distinguished basis.

    .. TODO::

        Extend this to have multiple actions, including actions on both sides.

    .. TODO::

        Extend when `M` does not have a basis and `S` is a permutation
        group using:

        - :arxiv:`0812.3082`
        - https://www.dmtcs.org/pdfpapers/dmAA0123.pdf
    """
    def __init__(self, M, S, action=operator.mul, side='left', *args, **kwargs):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(3)
            sage: R = G.regular_representation()
            sage: I = R.invariant_module()
            sage: TestSuite(I).run()

        TESTS::

            sage: G = GroupExp()(QQ) # a group that is not finitely generated
            sage: M = CombinatorialFreeModule(QQ, [1,2,3])
            sage: def on_basis(g, m): return M.monomial(m)  # trivial rep'n
            sage: from sage.modules.with_basis.representation import Representation
            sage: R = Representation(G, M, on_basis)
            sage: R.invariant_module()
            Traceback (most recent call last):
            ...
            ValueError: Multiplicative form of Rational Field is not finitely generated
        """
        if S not in FinitelyGeneratedSemigroups():
            raise ValueError(f"{S} is not finitely generated")
        if M not in FiniteDimensionalModulesWithBasis:
            raise ValueError(f"{M} is not a finite dimensional module with a distinguished basis")

        if side == "left":
            def _invariant_map(g, x):
                return action(g, x) - x
        elif side == "right":
            def _invariant_map(g, x):
                return action(x, g) - x
        else:
            raise ValueError("side must either be 'left' or 'right'")

        self._side = side
        self._action = action
        self._semigroup = S

        category = kwargs.pop("category", M.category().Subobjects())

        # Give the intersection of kernels of the map `s*x-x` to determine when
        # `s*x = x` for all generators `s` of `S`
        basis = M.annihilator_basis(S.gens(), action=_invariant_map, side='left')

        super().__init__(Family(basis),
                         support_order=M._compute_support_order(basis),
                         ambient=M,
                         unitriangular=False,
                         category=category,
                         *args, **kwargs)

    def construction(self):
        r"""
        Return the functorial construction of ``self``.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(3)
            sage: R = G.regular_representation(); R
            Left Regular Representation of Cyclic group of order 3 as a permutation group over Integer Ring
            sage: I = R.invariant_module()
            sage: I.construction()
            (EquivariantSubobjectConstructionFunctor,
            Left Regular Representation of Cyclic group of order 3 as a permutation group over Integer Ring)
        """
        from sage.categories.pushout import EquivariantSubobjectConstructionFunctor
        return (EquivariantSubobjectConstructionFunctor(self._semigroup,
                                                        self._action,
                                                        self._side),
                self.ambient())

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(3)
            sage: R = G.trivial_representation()
            sage: R.invariant_module()
            (Cyclic group of order 3 as a permutation group)-invariant submodule of
             Trivial representation of Cyclic group of order 3 as a permutation group over Integer Ring

            sage: G = CyclicPermutationGroup(3)
            sage: M = CombinatorialFreeModule(QQ, [1,2,3], prefix='M')
            sage: action = lambda g, m: M.monomial(g(m))  # cyclically permute coordinates
            sage: M.invariant_module(G, action_on_basis=action)
            (Cyclic group of order 3 as a permutation group)-invariant submodule of
             Free module generated by {1, 2, 3} over Rational Field
        """
        M = self._ambient
        if isinstance(self._ambient, Representation):
            M = M._module
        return f"({self._semigroup})-invariant submodule of {M}"

    def _latex_(self) -> str:
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(3)
            sage: R = G.algebra(QQ)
            sage: latex(R.invariant_module(G))
            \left( \Bold{Q}[\langle (1,2,3) \rangle] \right)^{\langle (1,2,3) \rangle}
        """
        M = self._ambient
        if isinstance(self._ambient, Representation):
            M = M._module
        from sage.misc.latex import latex
        return "\\left( {} \\right)^{{{}}}".format(latex(M), latex(self._semigroup))

    def _test_invariant(self, **options):
        """
        Check (on some elements) that ``self`` is invariant.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: G = SymmetricGroup(3)
            sage: M = CombinatorialFreeModule(QQ, [1,2,3], prefix='M')
            sage: def action(g, x): return M.monomial(g(x))
            sage: I = M.invariant_module(G, action_on_basis=action)
            sage: I._test_invariant()
            sage: G = SymmetricGroup(10)
            sage: M = CombinatorialFreeModule(QQ, list(range(1,11)), prefix='M')
            sage: def action(g, x): return M.monomial(g(x))
            sage: I = M.invariant_module(G, action_on_basis=action)
            sage: I._test_invariant(max_runs=10)
        """
        tester = self._tester(**options)
        X = tester.some_elements()
        L = []
        max_len = tester._max_runs

        # FIXME: This is max_len * dim number of runs!!!
        for i, x in enumerate(self._semigroup):
            L.append(x)
            if i >= max_len:
                break

        for x in L:
            for elt in X:
                lifted = self.lift(elt)
                if self._side == 'left':
                    tester.assertEqual(self._action(x, lifted), lifted)
                else:
                    tester.assertEqual(self._action(lifted, x), lifted)

    def semigroup(self):
        r"""
        Return the semigroup `S` whose action ``self`` is invariant under.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: G = SymmetricGroup(3)
            sage: M = CombinatorialFreeModule(QQ, [1,2,3], prefix='M')
            sage: def action(g, x): return M.monomial(g(x))
            sage: I = M.invariant_module(G, action_on_basis=action)
            sage: I.semigroup()
            Symmetric group of order 3! as a permutation group
        """
        return self._semigroup

    semigroup_representation = SubmoduleWithBasis.ambient

    class Element(SubmoduleWithBasis.Element):
        def _mul_(self, other):
            r"""
            Multiply ``self`` and ``other``.

            EXAMPLES:

            In general, there is not a well defined multiplication between
            two elements of a given module, but there is a multiplication
            with scalars::

                sage: M = CombinatorialFreeModule(QQ, [1,2,3], prefix='M');
                sage: G = CyclicPermutationGroup(3); G.rename('G')
                sage: g = G.an_element(); g
                (1,2,3)
                sage: from sage.modules.with_basis.representation import Representation
                sage: R = Representation(G,M,lambda g,x:M.monomial(g(x))); R.rename('R')
                sage: I = R.invariant_module()
                sage: B = I.basis()
                sage: [I.lift(b) for b in B]
                [M[1] + M[2] + M[3]]
                sage: v = B[0]
                sage: v*v
                Traceback (most recent call last):
                ...
                TypeError: unsupported operand parent(s) for *: 'R' and 'R'
                sage: (1/2) * v
                1/2*B[0]
                sage: v * (1/2)
                1/2*B[0]
                sage: R.rename()  # reset name

            Sometimes, the module is also a ring. To ensure the multiplication
            works as desired, we should be sure to pass the correct category to
            the :class:`~sage.modules.with_basis.representation.Representation`.
            In the following example, we use the exterior algebra over `\QQ`
            with three generators, which is in the category of finite
            dimensional `\QQ`-algebras with a basis::

                sage: G = CyclicPermutationGroup(3); G.rename('G')
                sage: M = algebras.Exterior(QQ, 'x', 3)
                sage: def on_basis(g, m): return M.prod([M.monomial(FrozenBitset([g(j+1)-1])) for j in m])  # cyclically permute generators
                sage: R = Representation(G, M, on_basis, category=Algebras(QQ).WithBasis().FiniteDimensional(), side='right')
                sage: I = R.invariant_module(); I.rename('I')
                sage: B = I.basis()
                sage: v = B[0] + 2*B[1]; I.lift(v)
                1 + 2*x0 + 2*x1 + 2*x2
                sage: w = B[2]; I.lift(w)
                x0*x1 - x0*x2 + x1*x2
                sage: v * w
                B[2] + 6*B[3]
                sage: I.lift(v*w)
                x0*x1 + 6*x0*x1*x2 - x0*x2 + x1*x2
                sage: w * v
                B[2] + 6*B[3]
                sage: (1/2) * v
                1/2*B[0] + B[1]
                sage: w * (1/2)
                1/2*B[2]
                sage: g = G((1,3,2))
                sage: v * g
                B[0] + 2*B[1]
                sage: w * g
                B[2]
                sage: g * v
                Traceback (most recent call last):
                ...
                TypeError: unsupported operand parent(s) for *: 'G' and 'I'
                sage: I.rename()  # reset name

                sage: R = Representation(G, M, on_basis, category=Algebras(QQ).WithBasis().FiniteDimensional())
                sage: I = R.invariant_module(); I.rename('I')
                sage: B = I.basis()
                sage: v = B[0] + 2*B[1]; I.lift(v)
                1 + 2*x0 + 2*x1 + 2*x2
                sage: w = B[2]; I.lift(w)
                x0*x1 - x0*x2 + x1*x2
                sage: v * w
                B[2] + 6*B[3]
                sage: I.lift(v*w)
                x0*x1 + 6*x0*x1*x2 - x0*x2 + x1*x2
                sage: w * v
                B[2] + 6*B[3]
                sage: (1/2) * v
                1/2*B[0] + B[1]
                sage: w * (1/2)
                1/2*B[2]
                sage: g = G((1,3,2))
                sage: v * v
                B[0] + 4*B[1]
                sage: g * w
                B[2]
                sage: v * g
                Traceback (most recent call last):
                ...
                TypeError: unsupported operand parent(s) for *: 'I' and 'G'
                sage: G.rename(); I.rename()  # reset names
            """
            P = self.parent()
            return P.retract(P.lift(self) * P.lift(other))

        def _acted_upon_(self, scalar, self_on_left=False):
            """
            EXAMPLES::

                sage: G = CyclicPermutationGroup(3)
                sage: g = G.an_element(); g
                (1,2,3)
                sage: M = CombinatorialFreeModule(QQ, [1,2,3])
                sage: from sage.modules.with_basis.representation import Representation
                sage: R = Representation(G, M, lambda g,x: M.monomial(g(x)))
                sage: I = R.invariant_module()
                sage: [b._acted_upon_(G((1,3,2))) for b in I.basis()]
                [B[0]]
                sage: v = I.an_element(); v
                2*B[0]
                sage: g * v
                2*B[0]
                sage: [g * v for g in G.list()]
                [2*B[0], 2*B[0], 2*B[0]]


                sage: E = algebras.Exterior(QQ, 'x', 3)
                sage: def on_basis(g, m): return E.prod([E.monomial(FrozenBitset([g(j+1)-1])) for j in m])  # cyclically permute generators
                sage: R = Representation(G, E, on_basis, category=Algebras(QQ).WithBasis().FiniteDimensional())
                sage: I = R.invariant_module()
                sage: B = I.basis()
                sage: [I.lift(b) for b in B]
                [1, x0 + x1 + x2, x0*x1 - x0*x2 + x1*x2, x0*x1*x2]
                sage: [[g*b for g in G] for b in B]
                [[B[0], B[0], B[0]],
                 [B[1], B[1], B[1]],
                 [B[2], B[2], B[2]],
                 [B[3], B[3], B[3]]]
                sage: 3 * I.basis()[0]
                3*B[0]
                sage: 3*B[0] + B[1]*2
                3*B[0] + 2*B[1]

                sage: R = G.regular_representation(QQ)
                sage: I = R.invariant_module()
                sage: B = I.basis()
                sage: [I.lift(b) for b in B]
                [() + (1,2,3) + (1,3,2)]
                sage: B[0]._acted_upon_(G((1,3,2)))
                B[0]
                sage: B[0]._acted_upon_(G((1,3,2)), self_on_left=True) is None
                True

                sage: R = G.regular_representation(QQ, side='right')
                sage: I = R.invariant_module()
                sage: B = I.basis()
                sage: [I.lift(b) for b in B]
                [() + (1,2,3) + (1,3,2)]
                sage: g = G((1,3,2))
                sage: B[0]._acted_upon_(g, self_on_left=True)
                B[0]
                sage: B[0]._acted_upon_(g, self_on_left=False) is None
                True

                sage: R = Representation(G, M, lambda g,x: M.monomial(g(x)), side='right')
                sage: I = R.invariant_module()
                sage: v = I.an_element(); v
                2*B[0]
                sage: v * g
                2*B[0]
                sage: [v * g for g in G.list()]
                [2*B[0], 2*B[0], 2*B[0]]
                sage: [b._acted_upon_(G((1,3,2)), self_on_left=True) for b in I.basis()]
                [B[0]]

                sage: def on_basis(g, m): return E.prod([E.monomial(FrozenBitset([g(j+1)-1])) for j in m])  # cyclically permute generators
                sage: R = Representation(G, E, on_basis, category=Algebras(QQ).WithBasis().FiniteDimensional(), side='right')
                sage: I = R.invariant_module()
                sage: B = I.basis()
                sage: [I.lift(b) for b in B]
                [1, x0 + x1 + x2, x0*x1 - x0*x2 + x1*x2, x0*x1*x2]
                sage: [[b * g for g in G] for b in B]
                [[B[0], B[0], B[0]],
                 [B[1], B[1], B[1]],
                 [B[2], B[2], B[2]],
                 [B[3], B[3], B[3]]]
                sage: 3 * B[0] + B[1] * 2
                3*B[0] + 2*B[1]
            """
            if scalar in self.parent()._semigroup and self_on_left == (self.parent()._side == 'right'):
                return self
            return super()._acted_upon_(scalar, self_on_left)


class FiniteDimensionalTwistedInvariantModule(SubmoduleWithBasis):
    r"""
    Construct the `\chi`-twisted invariant submodule of `M`.

    When a group `G` acts on a module `M`, the `\chi`-*twisted invariant
    submodule* of `M` is the isotypic component of the representation `M`
    corresponding to the irreducible character `\chi`.

    For more information, see [Sta1979]_.

    INPUT:

    - ``M`` -- a module in the category of
      :class:`~sage.categories.finite_dimensional_modules_with_basis.FiniteDimensionalModulesWithBasis`
      and whose base ring contains all the values passed to ``chi`` and `1/|G|`

    - ``G`` -- a finitely generated group

    - ``chi`` -- list/tuple of the character values of the irreducible representation
      onto which you want to project. The order of values of `chi` must
      agree with the order of ``G.conjugacy_classes()``

    - ``action`` -- (default: ``operator.mul``) the action of ``G`` on ``M``

    - ``side`` -- (default: ``'left'``) the side on which ``G`` acts

    .. WARNING::

        The current implementation does not check if ``chi`` is irreducible.
        Passing character values of non-irreducible representations may lead
        to mathematically incorrect results.

    EXAMPLES:

    Suppose that the symmetric group `S_3` acts on a four dimensional
    vector space by permuting the first three coordinates only::

        sage: M = CombinatorialFreeModule(QQ, [1,2,3,4], prefix='M')
        sage: G = SymmetricGroup(3)
        sage: action = lambda g,x: M.term(g(x))

    The trivial representation corresponds to the usual invariant module,
    so trying to create the twisted invariant module when there is no twist
    returns a :class:`~sage.modules.with_basis.invariant.FiniteDimensionalInvariantModule`::

        sage: # needs sage.rings.number_field
        sage: chi = ClassFunction(G, (1,1,1))
        sage: T = M.twisted_invariant_module(G, chi, action_on_basis=action)
        sage: type(T)
        <class 'sage.modules.with_basis.invariant.FiniteDimensionalInvariantModule_with_category'>

    In this case, there are two copies of the trivial representation, one
    coming from the first three coordinates and the other coming from the
    fact that `S_3` does not touch the fourth coordinate::

        sage: # needs sage.rings.number_field
        sage: T.basis()
        Finite family {0: B[0], 1: B[1]}
        sage: [T.lift(b) for b in T.basis()]
        [M[1] + M[2] + M[3], M[4]]

    The character values of the standard representation are `2,0,-1`::

        sage: # needs sage.rings.number_field
        sage: chi = ClassFunction(G, [2,0,-1])
        sage: T = M.twisted_invariant_module(G, chi, action_on_basis=action)
        sage: type(T)
        <class 'sage.modules.with_basis.invariant.FiniteDimensionalTwistedInvariantModule_with_category'>
        sage: T.basis()
        Finite family {0: B[0], 1: B[1]}
        sage: [T.lift(b) for b in T.basis()]
        [M[1] - M[3], M[2] - M[3]]

    The permutation representation is the direct sum of the standard
    representation with the trivial representation, and the action on the
    basis element ``B[4]`` is itself a copy of the trivial representation,
    so the sign representation does not appear in the decomposition::

        sage: # needs sage.rings.number_field
        sage: T = M.twisted_invariant_module(G, [1,-1,1], action_on_basis=action)
        sage: T.basis()
        Finite family {}

    We can also get two copies of the standard representation by looking at
    two copies of the permutation representation, found by reduction modulo
    three on the indices of a six-dimensional module::

        sage: M = CombinatorialFreeModule(QQ, [0,1,2,3,4,5], prefix='M')
        sage: action = lambda g,x: M.term(g(x%3 + 1)-1 + (x>=3)*3)
        sage: T = M.twisted_invariant_module(G, [2,0,-1], action_on_basis=action)
        sage: T.basis()
        Finite family {0: B[0], 1: B[1], 2: B[2], 3: B[3]}
        sage: [T.lift(b) for b in T.basis()]
        [M[0] - M[2], M[1] - M[2], M[3] - M[5], M[4] - M[5]]

        sage: T = M.twisted_invariant_module(G, [1,1,1], action_on_basis=action)
        sage: T.basis()
        Finite family {0: B[0], 1: B[1]}
        sage: [T.lift(b) for b in T.basis()]
        [M[0] + M[1] + M[2], M[3] + M[4] + M[5]]

    There are still no copies of the sign representation::

        sage: T = M.twisted_invariant_module(G, [1,-1,1], action_on_basis=action)
        sage: T.basis()
        Finite family {}

    The trivial representation also contains no copies of the sign
    representation::

        sage: R = G.trivial_representation(QQ)
        sage: T = R.twisted_invariant_module([1,-1,1])
        sage: T.basis()
        Finite family {}

    The regular representation contains two copies of the standard
    representation and one copy each of the trivial and the sign::

        sage: R = G.regular_representation(QQ)
        sage: std = R.twisted_invariant_module([2,0,-1])
        sage: std.basis()
        Finite family {0: B[0], 1: B[1], 2: B[2], 3: B[3]}
        sage: [std.lift(b) for b in std.basis()]
        [() - (1,2,3), -(1,2,3) + (1,3,2), (2,3) - (1,2), -(1,2) + (1,3)]

        sage: triv = R.twisted_invariant_module([1,1,1])
        sage: triv.basis()
        Finite family {0: B[0]}
        sage: [triv.lift(b) for b in triv.basis()]
        [() + (2,3) + (1,2) + (1,2,3) + (1,3,2) + (1,3)]

        sage: sgn = R.twisted_invariant_module([1,-1,1])
        sage: sgn.basis()
        Finite family {0: B[0]}
        sage: [sgn.lift(b) for b in sgn.basis()]
        [() - (2,3) - (1,2) + (1,2,3) + (1,3,2) - (1,3)]

    For the next example, we construct a twisted invariant by the character
    for the 2 dimensional representation of `S_3` on the natural action on
    the exterior algebra. While `S_3` acts by automorphisms, the twisted
    invariants do not form an algebra in this case::

        sage: G = SymmetricGroup(3); G.rename('S3')
        sage: E = algebras.Exterior(QQ, 'x', 3); E.rename('E')
        sage: def action(g, m): return E.prod([E.monomial(FrozenBitset([g(j+1)-1])) for j in m])
        sage: from sage.modules.with_basis.representation import Representation
        sage: EA = Representation(G, E, action, category=Algebras(QQ).WithBasis().FiniteDimensional())
        sage: T = EA.twisted_invariant_module([2,0,-1])
        sage: t = T.an_element(); t
        2*B[0] + 2*B[1] + 3*B[2]

    We can still get meaningful information about the product
    by taking the product in the ambient space::

        sage: T.lift(t) * T.lift(t)
        -36*x0*x1*x2

    We can see this does not lie in this twisted invariant algebra::

        sage: T.retract(T.lift(t) * T.lift(t))
        Traceback (most recent call last):
        ...
        ValueError: -36*x0*x1*x2 is not in the image

        sage: [T.lift(b) for b in T.basis()]
        [x0 - x2, x1 - x2, x0*x1 - x1*x2, x0*x2 + x1*x2]

    It happens to be in the trivial isotypic component (equivalently in
    the usual invariant algebra) but Sage does not know this.

    ::

        sage: G.rename(); E.rename()  # reset the names

    .. TODO::

        - Replace ``G`` by ``S`` in :class:`~sage.categories.finitely_generated_semigroups.FinitelyGeneratedSemigroups`
        - Allow for ``chi`` to be a :class:`~sage.modules.with_basis.representation.Representation`
        - Add check for irreducibility of ``chi``
    """

    @staticmethod
    def __classcall_private__(cls, M, G, chi,
                              action=operator.mul, side='left', **kwargs):
        r"""
        TESTS:

        Check that it works for lists::

            sage: M = CombinatorialFreeModule(QQ, [1,2,3])
            sage: G = SymmetricGroup(3)
            sage: def action(g, x): return M.term(g(x))
            sage: T = M.twisted_invariant_module(G, [2,0,-1], action_on_basis=action)   # needs sage.combinat

        Check that it works for tuples::

            sage: T2 = M.twisted_invariant_module(G, (2,0,-1), action_on_basis=action)  # needs sage.combinat
            sage: T is T2                                                               # needs sage.combinat
            True

        Check that it works for class functions::

            sage: chi = ClassFunction(G, [2,0,-1])
            sage: T3 = M.twisted_invariant_module(G, chi, action_on_basis=action)       # needs sage.combinat
            sage: T is T3                                                               # needs sage.combinat
            True

        Check that it works when the character values are not an instance of
        :class:`~sage.rings.integer.Integer`::

            sage: chi = [QQ(2), QQ(0), QQ(-1)]
            sage: T4 = M.twisted_invariant_module(G, chi, action_on_basis=action)       # needs sage.combinat
            sage: T is T4                                                               # needs sage.combinat
            True

        Check that the trivial character returns an instance of
        :class:`~sage.modules.with_basis.invariant.FiniteDimensionalInvariantModule`::

            sage: # needs sage.combinat
            sage: chi = [1, 1, 1] # check for list
            sage: T = M.twisted_invariant_module(G, chi, action_on_basis=action)
            sage: type(T)
            <class 'sage.modules.with_basis.invariant.FiniteDimensionalInvariantModule_with_category'>
            sage: chi = (1, 1, 1) # check for tuple
            sage: T = M.twisted_invariant_module(G, chi, action_on_basis=action)
            sage: type(T)
            <class 'sage.modules.with_basis.invariant.FiniteDimensionalInvariantModule_with_category'>
            sage: chi = ClassFunction(G, [1,1,1]) # check for class function
            sage: T = M.twisted_invariant_module(G, chi, action_on_basis=action)
            sage: type(T)
            <class 'sage.modules.with_basis.invariant.FiniteDimensionalInvariantModule_with_category'>

        Check the :exc:`ValueError`::

            sage: T = M.twisted_invariant_module(G, "ichigo", action_on_basis=action)   # needs sage.combinat
            Traceback (most recent call last):
            ...
            ValueError: chi must be a list/tuple or a class function of the group G
        """
        from sage.groups.class_function import ClassFunction, ClassFunction_libgap
        from sage.libs.gap.libgap import libgap

        if isinstance(chi, (list, tuple)):
            chi = ClassFunction(G, libgap(chi))
        elif not isinstance(chi, ClassFunction_libgap):
            raise ValueError("chi must be a list/tuple or a class function of the group G")

        try:
            is_trivial = all(chi(next(iter(conj))) == 1 for conj in G.conjugacy_classes())
        except AttributeError:  # to handle ReflectionGroups
            is_trivial = all(chi(G(next(iter(conj)))) == 1 for conj in G.conjugacy_classes())

        if is_trivial:
            action_on_basis = kwargs.pop('action_on_basis', None)
            if action_on_basis is not None:
                return M.invariant_module(G, action_on_basis=action_on_basis)
            return M.invariant_module(G, action=action)

        return super().__classcall__(cls, M, G, chi, action=operator.mul,
                                     side='left', **kwargs)

    def __init__(self, M, G, chi, action=operator.mul, side='left', **kwargs):
        r"""
        Initialize ``self``.

        EXAMPLES:

        As a first example we will consider the permutation representation
        of `S_3`::

            sage: M = CombinatorialFreeModule(QQ, [1,2,3], prefix='M');
            sage: G = SymmetricGroup(3)
            sage: G.conjugacy_classes()                                                 # needs sage.libs.flint
            [Conjugacy class of cycle type [1, 1, 1] in Symmetric group of order 3! as a permutation group,
             Conjugacy class of cycle type [2, 1] in Symmetric group of order 3! as a permutation group,
             Conjugacy class of cycle type [3] in Symmetric group of order 3! as a permutation group]
            sage: from sage.groups.class_function import ClassFunction
            sage: chi = ClassFunction(G, [2,0,-1]) # the standard representation character values
            sage: def action(g, x): return M.term(g(x))
            sage: import __main__
            sage: __main__.action = action
            sage: T = M.twisted_invariant_module(G, chi, action_on_basis=action)
            sage: TestSuite(T).run()

        We know that the permutation representation decomposes as a direct
        sum of one copy of the standard representation which is two-dimensional
        and one copy of the trivial representation::

            sage: T.basis()
            Finite family {0: B[0], 1: B[1]}
            sage: [T.lift(b) for b in T.basis()]
            [M[1] - M[3], M[2] - M[3]]
        """

        if G not in FinitelyGeneratedSemigroups():
            raise ValueError(f"{G} is not finitely generated")
        if M not in FiniteDimensionalModulesWithBasis:
            raise ValueError(f"{M} is not a finite dimensional module with a distinguished basis")

        self._chi = chi
        self._group = G
        self._action = action
        self._side = side

        # define a private action to deal
        # with sidedness issues in the action.
        if side == 'left':
            self.__sided_action__ = action
        elif side == 'right':
            # flip the sides since the second argument
            # to action should be the group element
            def __sided_action__(g, x):
                return action(x, g)
            self.__sided_action__ = __sided_action__
        else:
            raise ValueError("side must either be 'left' or 'right'")

        proj_matrix = Matrix(M.dimension())  # initialize the zero-matrix
        for g in self._group:
            proj_matrix += self._chi(g)*Matrix((self.__sided_action__(g, b)).to_vector() for b in M.basis())

        n = self._chi(self._group.identity())  # chi(1) is the dimension
        g = self._group.order()

        self._projection_matrix = (n/g)*proj_matrix

        self._project_ambient = M.module_morphism(matrix=self._projection_matrix,
                                                  codomain=M)

        category = kwargs.pop("category", M.category().Subobjects())

        # Give the kernel of the map `\pi(x)-x` to determine when `x` lies
        # within the isotypic component of `R`.

        def proj_difference(g, x):
            return self._project_ambient(x) - x

        basis = M.annihilator_basis(M.basis(),
                                    action=proj_difference,
                                    side='left')

        super().__init__(Family(basis),
                         support_order=M._compute_support_order(basis),
                         ambient=M,
                         unitriangular=False,
                         category=category,
                         **kwargs)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(3)
            sage: M = CombinatorialFreeModule(QQ, [1,2,3], prefix='M')
            sage: action = lambda g, m: M.monomial(g(m))  # cyclically permute coordinates
            sage: M.twisted_invariant_module(G, [2,0,-1], action_on_basis=action)
            Twist of (Cyclic group of order 3 as a permutation group)-invariant submodule of
             Free module generated by {1, 2, 3} over Rational Field by character [2, 0, -1]
        """
        M = self._ambient
        if isinstance(self._ambient, Representation):
            M = M._module
        return f"Twist of ({self._group})-invariant submodule of {M} by character {self._chi.values()}"

    def project(self, x):
        r"""
        Project ``x`` in the ambient module onto ``self``.

        EXAMPLES:

        The standard representation is the orthogonal complement
        of the trivial representation inside of the permutation
        representation, so the basis for the trivial representation
        projects to `0`::

            sage: M = CombinatorialFreeModule(QQ, [1,2,3]); M.rename('M')
            sage: B = M.basis()
            sage: G = SymmetricGroup(3); G.rename('S3')
            sage: def action(g, x): return M.term(g(x))
            sage: T = M.twisted_invariant_module(G, [2,0,-1], action_on_basis=action)
            sage: m = B[1] + B[2] + B[3]
            sage: parent(m)
            M
            sage: t = T.project(m); t
            0
            sage: parent(t)
            Twist of (S3)-invariant submodule of M by character [2, 0, -1]

            sage: G.rename(); M.rename()  # reset names
        """
        return self.retract(self.project_ambient(x))

    def project_ambient(self, x):
        r"""
        Project ``x`` in the ambient representation onto the submodule of the
        ambient representation to which ``self`` is isomorphic as a module.

        .. NOTE::

            The image of ``self.project_ambient`` is not in ``self`` but
            rather is in ``self.ambient()``.

        EXAMPLES::

            sage: M = CombinatorialFreeModule(QQ, [1,2,3]); M.rename('M')
            sage: B = M.basis()
            sage: G = SymmetricGroup(3); G.rename('S3')
            sage: def action(g, x): return M.term(g(x))
            sage: T = M.twisted_invariant_module(G, [2,0,-1], action_on_basis=action)

        To compare with ``self.project``, we can inspect the parents.
        The image of ``self.project`` is in ``self``, while the image
        of ``self.project_ambient`` is in ``self._ambient``::

            sage: t = T.project(B[1] + B[2] + B[3]); t
            0
            sage: parent(t)
            Twist of (S3)-invariant submodule of M by character [2, 0, -1]
            sage: s = T.project_ambient(B[1] + B[2] + B[3]); s
            0
            sage: parent(s)
            Representation of S3 indexed by {1, 2, 3} over Rational Field

        Note that because of the construction of ``T``, ``self._ambient``
        is an instance of
        :class:`~sage.modules.with_basis.representation.Representation`,
        but you still may pass elements of ``M``, which is an instance of
        :class:`~sage.combinat.free_module.CombinatorialFreeModule`,
        because the underlying ``Representation`` is built off of ``M``
        and we can canonically construct elements of the ``Representation``
        from elements of ``M``.

        ::

            sage: G.rename(); M.rename()  # reset names
        """
        if (isinstance(self._ambient, Representation)
                and x.parent() is self._ambient._module):
            x = self._ambient._element_constructor_(x)
        return self._project_ambient(x)

    def projection_matrix(self):
        r"""
        Return the matrix defining the projection map from
        the ambient representation onto ``self``.

        EXAMPLES::

            sage: M = CombinatorialFreeModule(QQ, [1,2,3])
            sage: def action(g, x): return(M.term(g(x)))
            sage: G = SymmetricGroup(3)

        If the matrix `A` has columns form a basis for
        the subspace onto which we are trying to project,
        then we can find the projection matrix via the
        formula `P = A (A^T A)^{-1} A^T`. Recall that the
        standard representation twisted invariant has basis
        ``(B[1] - B[3], B[2] - B[3])``, hence::

            sage: A = Matrix([[1,0],[0,1],[-1,-1]])
            sage: P = A*(A.transpose()*A).inverse()*A.transpose()
            sage: T = M.twisted_invariant_module(G, [2,0,-1], action_on_basis=action)
            sage: P == T.projection_matrix()
            True

        Moreover, since there is no component of the sign
        representation in this representation, the projection
        matrix is just the zero matrix::

            sage: T = M.twisted_invariant_module(G, [1,-1,1], action_on_basis=action)
            sage: T.projection_matrix()
            [0 0 0]
            [0 0 0]
            [0 0 0]
        """
        return self._projection_matrix

    class Element(SubmoduleWithBasis.Element):
        def _acted_upon_(self, scalar, self_on_left=False):
            r"""
            Return the action of ``scalar`` on ``self``.

            EXAMPLES::

                sage: G = SymmetricGroup(3)
                sage: R = G.regular_representation(QQ)
                sage: T = R.twisted_invariant_module([2,0,-1])
                sage: t = T.an_element()
                sage: 5 * t
                10*B[0] + 10*B[1] + 15*B[2]
                sage: t * -2/3
                -4/3*B[0] - 4/3*B[1] - 2*B[2]
                sage: [g * t for g in G]
                [2*B[0] + 2*B[1] + 3*B[2],
                 -4*B[0] + 2*B[1] - 3*B[3],
                 2*B[0] - 4*B[1] - 3*B[2] + 3*B[3],
                 3*B[0] + 2*B[2] + 2*B[3],
                 -3*B[1] - 4*B[2] + 2*B[3],
                 -3*B[0] + 3*B[1] + 2*B[2] - 4*B[3]]
            """
            P = self.parent()
            if scalar in P._group and self_on_left == (P._side == 'right'):
                return P.retract(scalar * P.lift(self))
            return super()._acted_upon_(scalar, self_on_left)
